from __future__ import annotations

import copy
import keyword
import re
import sys
import unicodedata
from urllib.parse import urlparse

from packaging.version import InvalidVersion, Version

if sys.version_info >= (3, 11):  # pragma: no-cover-if-lt-py311
    import tomllib
else:  # pragma: no-cover-if-gte-py311
    import tomli as tomllib

from briefcase.debuggers.base import BaseDebugger
from briefcase.platforms import get_output_formats, get_platforms

from .constants import RESERVED_WORDS
from .exceptions import BriefcaseConfigError, InvalidVersionError

# PEP 508 restricts the naming of modules. The PEP defines a regex that uses
# re.IGNORECASE; but in in practice, packaging uses a version that rolls out the lower
# case, which has very slightly different semantics with non-ASCII characters. This
# definition is the one from
# https://github.com/pypa/packaging/blob/24.0/src/packaging/_tokenizer.py#L80
PEP508_NAME_RE = re.compile(r"^([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9._-]*[a-zA-Z0-9])$")


def is_valid_pep508_name(app_name):
    """Determine if the name is valid by PEP508 rules."""
    return PEP508_NAME_RE.match(app_name)


def is_reserved_keyword(app_name):
    """Determine if the name is a reserved keyword."""
    return keyword.iskeyword(app_name.lower()) or app_name.lower() in RESERVED_WORDS


def is_valid_app_name(app_name):
    return not is_reserved_keyword(app_name) and is_valid_pep508_name(app_name)


def make_class_name(formal_name):
    """Construct a valid class name from a formal name.

    :param formal_name: The formal name
    :returns: The app's class name
    """
    # Identifiers (including class names) can be unicode.
    # https://docs.python.org/3/reference/lexical_analysis.html#identifiers
    xid_start = {
        "Lu",  # uppercase letters
        "Ll",  # lowercase letters
        "Lt",  # titlecase letters
        "Lm",  # modifier letters
        "Lo",  # other letters
        "Nl",  # letter numbers
    }
    xid_continue = xid_start.union(
        {
            "Mn",  # nonspacing marks
            "Mc",  # spacing combining marks
            "Nd",  # decimal number
            "Pc",  # connector punctuation
        }
    )

    # Normalize to NFKC form, then remove any character that isn't
    # in the allowed categories, or is the underscore character;
    # Capitalize the resulting word.
    class_name = "".join(
        ch
        for ch in unicodedata.normalize("NFKC", formal_name)
        if unicodedata.category(ch) in xid_continue or ch == "_"
    )

    # If the first character isn't in the 'start' character set,
    # and it isn't already an underscore, prepend an underscore.
    if (
        class_name
        and unicodedata.category(class_name[0]) not in xid_start
        and class_name[0] != "_"
    ):
        class_name = f"_{class_name}"

    return class_name


def validate_url(candidate):
    """Determine if the URL is valid.

    :param candidate: The candidate URL
    :returns: True. If there are any validation problems, raises ValueError with a
        diagnostic message.
    """
    result = urlparse(candidate)
    if not all([result.scheme, result.netloc]):
        raise ValueError("Not a valid URL!")
    if result.scheme not in {"http", "https"}:
        raise ValueError("Not a valid website URL!")
    return True


def validate_document_type_config(document_type_id, document_type):
    try:
        if not (
            isinstance(document_type["extension"], str)
            and document_type["extension"].isalnum()
        ):
            raise BriefcaseConfigError(
                f"The extension provided for document type "
                f"{document_type_id!r} is not alphanumeric."
            )
    except KeyError:
        raise BriefcaseConfigError(
            f"Document type {document_type_id!r} does not provide an extension."
        ) from None

    try:
        if not isinstance(document_type["icon"], str):
            raise BriefcaseConfigError(
                f"The icon definition associated with document type "
                f"{document_type_id!r} is not a string."
            )
    except KeyError:
        raise BriefcaseConfigError(
            f"Document type {document_type_id!r} does not define an icon."
        ) from None

    try:
        if not isinstance(document_type["description"], str):
            raise BriefcaseConfigError(
                f"The description associated with document type "
                f"{document_type_id!r} is not a string."
            )
    except KeyError:
        raise BriefcaseConfigError(
            f"Document type {document_type_id!r} does not provide a description."
        ) from None

    try:
        validate_url(document_type["url"])
    except KeyError:
        raise BriefcaseConfigError(
            f"Document type {document_type_id!r} does not provide a URL."
        ) from None
    except ValueError as e:
        raise BriefcaseConfigError(
            f"The URL associated with document type {document_type_id!r} "
            f"is invalid: {e}"
        ) from None

    if sys.platform == "darwin":  # pragma: no-cover-if-not-macos
        from briefcase.platforms.macOS.utils import is_uti_core_type, mime_type_to_uti

        macOS = document_type.setdefault("macOS", {})
        content_types = macOS.get("LSItemContentTypes", None)
        mime_type = document_type.get("mime_type", None)

        if isinstance(content_types, list):
            if len(content_types) > 1:
                raise BriefcaseConfigError(
                    f"""
Document type {document_type_id!r} has multiple content types. Specifying
multiple values in a LSItemContentTypes key is only valid when multiple document
types are manually grouped together in the Info.plist file. For Briefcase apps,
document types are always separately declared in the configuration file, so only
a single value should be provided.
                """
                )

            macOS["LSItemContentTypes"] = content_types
            uti = content_types[0]
        elif isinstance(content_types, str):
            # If the content type is a string, convert it to a list
            macOS["LSItemContentTypes"] = [content_types]
            uti = content_types
        else:
            uti = None

        # If an UTI is provided in LSItemContentTypes,
        # that takes precedence over a MIME type
        if is_uti_core_type(uti) or ((uti := mime_type_to_uti(mime_type)) is not None):
            macOS.setdefault("is_core_type", True)
            macOS.setdefault("LSItemContentTypes", [uti])
            macOS.setdefault("LSHandlerRank", "Alternate")
        else:
            # LSItemContentTypes will default to bundle.app_name.document_type_id
            # in the Info.plist template if it is not provided.
            macOS.setdefault("is_core_type", False)
            macOS.setdefault("LSHandlerRank", "Owner")
            macOS.setdefault("UTTypeConformsTo", ["public.data", "public.content"])

        macOS.setdefault("CFBundleTypeRole", "Viewer")
    else:  # pragma: no-cover-if-is-macos
        pass


def validate_install_options_config(config, opt_type, **others):
    """Validate that install/uninstall options are valid and complete, and convert to a
    dict.

    The dict format is required because Cookiecutter doesn't allow passing a list as a
    context value; you have to use the reliable iteration order of a dict instead.

    :param config: The table form of options
    :param opt_type: The label of the option type being parsed ("install" or
        "uninstall")
    :param others: A dictionary of other parsed option types. The keys are
        the option types, and the values are the dictionary of parse options. Options
        in `config` must be unique against these keys.
    """
    options = {}
    known_names = set()
    if config:
        for i, config_item in enumerate(config):
            try:
                name = config_item["name"]
                if not isinstance(name, str):
                    raise BriefcaseConfigError(
                        f"Name for {opt_type} option {i} is not a string."
                    )
            except KeyError:
                raise BriefcaseConfigError(
                    f"{opt_type.title()} option {i} does not define a `name`."
                ) from None

            # Options must be valid Python identifiers
            if not name.isidentifier():
                raise BriefcaseConfigError(
                    f"{name!r} cannot be used as an {opt_type} option name, "
                    "as it is not a valid Python identifier."
                )

            # Option names may be coerced into upper case; and there are
            # a small number of reserved identifiers.
            if name.upper() == "ALLUSERS":
                raise BriefcaseConfigError(
                    f"{name!r} is a reserved {opt_type} option identifier."
                )

            option = {}
            if name.upper() in known_names:
                raise BriefcaseConfigError(
                    f"{opt_type.title()} option names must be unique. The name "
                    f"{name!r}, used by {opt_type} option {i}, has already "
                    "been defined."
                )
            else:
                for other_type, other_options in others.items():
                    if name.upper() in {n.upper() for n in other_options}:
                        raise BriefcaseConfigError(
                            f"{opt_type.title()} option names must be unique. The name "
                            f"{name!r} is already used as an {other_type} option."
                        )

            # options needs to retain the original name, but we need names to be
            # case-unique as well, so we track a separate set of known upper case names.
            known_names.add(name.upper())
            options[name] = option

            try:
                # Options must have a string title.
                option["title"] = config_item["title"]
                if not isinstance(option["title"], str):
                    raise BriefcaseConfigError(
                        f"Title for {opt_type} option {name!r} is not a string."
                    )
            except KeyError:
                raise BriefcaseConfigError(
                    f"{opt_type.title()} option {name!r} does not provide a title."
                ) from None

            try:
                # Options must have a string title.
                option["description"] = config_item["description"]
                if not isinstance(option["description"], str):
                    raise BriefcaseConfigError(
                        f"Description for {opt_type} option {name!r} is not a string."
                    )
            except KeyError:
                raise BriefcaseConfigError(
                    f"{opt_type.title()} option {name!r} does not provide "
                    "a description."
                ) from None

            # Options are booleans, and are False by default
            option["default"] = bool(config_item.get("default", False))

    return options


VALID_BUNDLE_RE = re.compile(r"[a-zA-Z0-9-]+(\.[a-zA-Z0-9-]+)+$")


def is_valid_bundle_identifier(bundle):
    """Check if the bundle identifier follows the basic reversed domain name pattern."""
    return VALID_BUNDLE_RE.match(bundle) is not None


def parse_boolean(value: str) -> bool:
    """Takes a string value and attempts to convert to a boolean value."""

    truth_vals = {"true", "t", "yes", "y", "1", "on"}
    false_vals = {"false", "f", "no", "n", "0", "off"}

    normalised_val = value.strip().lower()

    if normalised_val in truth_vals:
        return True
    elif normalised_val in false_vals:
        return False
    else:
        raise ValueError(
            f"Invalid boolean value: {value!r}. "
            f"Expected one of {truth_vals | false_vals}"
        )


class BaseConfig:
    def __init__(self, **kwargs):
        for attr, value in kwargs.items():
            setattr(self, attr, value)

    def update(self, data):
        """Add fields and values from data to BaseConfig instance.

        Any existing fields named in data will be overwritten. To merge
        data with existing configs, use the `merge_config` function.

        :param data: The new configuration data dictionary.
        """
        for key, configs in data.items():
            setattr(self, key, configs)

    def copy(self):
        return type(self)(**self.__dict__)

    def setdefault(self, field_name, default_value):
        """Return the field_name field or, if it does not exist, create it to hold
        default_value.

        Behaves similarly to dict.setdefault().

        :param field_name: The name of the desired/new field.
        :param default_value: The value to assign to self.field_name if it does not
            already exist.
        """
        if not hasattr(self, field_name):
            setattr(self, field_name, default_value)

        return getattr(self, field_name)


class GlobalConfig(BaseConfig):
    def __init__(
        self,
        project_name,
        version,
        bundle,
        license=None,
        url=None,
        author=None,
        author_email=None,
        requires_python=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.project_name = project_name
        self.version = version
        self.bundle = bundle
        self.url = url
        self.author = author
        self.author_email = author_email
        self.license = license
        self.requires_python = requires_python

        # Version number is compliant with PEP440 (and related updates):
        try:
            # If input is already a version object (can happen by copying), use as-is
            if isinstance(version, Version):
                self.version = version
            else:
                self.version = Version(version)
        except InvalidVersion:
            raise InvalidVersionError(
                f"Version number ({self.version}) is not valid."
            ) from None

    def __repr__(self):
        return f"<{self.project_name} v{self.version} GlobalConfig>"


class AppConfig(BaseConfig):
    def __init__(
        self,
        app_name,
        version,
        bundle,
        description,
        license,
        sources=None,
        formal_name=None,
        url=None,
        author=None,
        author_email=None,
        requires=None,
        icon=None,
        document_type=None,
        install_option=None,
        uninstall_option=None,
        permission=None,
        template=None,
        template_branch=None,
        test_sources=None,
        test_requires=None,
        supported=True,
        long_description=None,
        console_app=False,
        requirement_installer_args: list[str] | None = None,
        external_package_path: str | None = None,
        external_package_executable_path: str | None = None,
        install_launcher: bool | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        # All app configs are created in unfinalized draft form.
        self.__draft__ = True

        self.app_name = app_name
        self.version = version
        self.bundle = bundle.lower()
        # Description can only be a single line. Ignore everything else.
        self.description = description.split("\n")[0]
        self.sources = sources
        self.formal_name = app_name if formal_name is None else formal_name
        self.url = url
        self.author = author
        self.author_email = author_email
        self.requires = requires
        self.icon = icon
        self.document_types = {} if document_type is None else document_type
        self.permission = {} if permission is None else permission
        self.template = template
        self.template_branch = template_branch
        self.test_sources = test_sources
        self.test_requires = test_requires
        self.supported = supported
        self.long_description = long_description
        self.license = license
        self.console_app = console_app
        self.requirement_installer_args = (
            [] if requirement_installer_args is None else requirement_installer_args
        )
        self.external_package_path = external_package_path
        self.external_package_executable_path = external_package_executable_path
        self.install_launcher = (
            install_launcher if (install_launcher is not None) else (not console_app)
        )

        self.test_mode: bool = False

        self.debugger: BaseDebugger | None = None
        self.debugger_host: str | None = None  # only for run command
        self.debugger_port: int | None = None  # only for run command

        if not is_valid_app_name(self.app_name):
            raise BriefcaseConfigError(
                f"{self.app_name!r} is not a valid app name."
                f"\n\n"
                "App names must not be reserved keywords such as 'and', 'for' and "
                "'while'. They must also be PEP508 compliant (i.e., they can only "
                "include letters, numbers, '-' and '_'; must start with a letter; "
                "and cannot end with '-' or '_')."
            )

        if not is_valid_bundle_identifier(self.bundle):
            raise BriefcaseConfigError(
                f"{self.bundle!r} is not a valid bundle identifier."
                f"\n\n"
                "The bundle should be a reversed domain name. It must contain at least "
                "2 dot-separated sections; each section may only include letters, "
                "numbers, and hyphens; and each section may not contain any reserved "
                "words (like 'switch', or 'while')."
            )

        for document_type_id, document_type in self.document_types.items():
            validate_document_type_config(document_type_id, document_type)

        self.install_options = validate_install_options_config(
            install_option, "install"
        )
        self.uninstall_options = validate_install_options_config(
            uninstall_option,
            "uininstall",
            install=self.install_options,
        )

        # Version number is compliant with PEP440 (and related updates):
        try:
            # If input is already a version object (can happen by copying), use as-is
            if isinstance(version, Version):
                self.version = version
            else:
                self.version = Version(version)
        except InvalidVersion:
            raise InvalidVersionError(
                f"Version number for {self.app_name!r} ({self.version}) is not valid."
            ) from None

        if self.sources:
            # Sources list doesn't include any duplicates
            source_modules = {source.rsplit("/", 1)[-1] for source in self.sources}
            if len(self.sources) != len(source_modules):
                raise BriefcaseConfigError(
                    f"The `sources` list for {self.app_name!r} contains duplicated "
                    "package names."
                )

            # There is, at least, a source for the app module
            if self.module_name not in source_modules:
                raise BriefcaseConfigError(
                    f"The `sources` list for {self.app_name!r} does not include a "
                    f"package named {self.module_name!r}."
                )

    def __repr__(self):
        return f"<{self.bundle_identifier} v{self.version} AppConfig>"

    @property
    def module_name(self):
        """The module name for the app.

        This is derived from the name, but:
        * all `-` have been replaced with `_`.
        """
        return self.app_name.replace("-", "_")

    @property
    def bundle_name(self):
        """The bundle name for the app.

        This is derived from the app name, but:
        * all `_` have been replaced with `-`.
        """
        return self.app_name.replace("_", "-").lower()

    @property
    def bundle_identifier(self):
        """The bundle identifier for the app.

        This is derived from the bundle and the bundle name, joined by a `.`.
        """
        return f"{self.bundle}.{self.bundle_name}"

    @property
    def class_name(self):
        """The class name for the app.

        This is derived from the formal name for the app.
        """
        return make_class_name(self.formal_name)

    @property
    def package_name(self):
        """The bundle name of the app, with `-` replaced with `_` to create something
        that can be used a namespace identifier on Python or Java, similar to
        `module_name`."""
        return self.bundle.replace("-", "_")

    @property
    def dist_info_name(self):
        """The name of the .dist-info directory for the app."""
        return f"{self.module_name}.dist-info"

    def PYTHONPATH(self):
        """The PYTHONPATH modifications needed to run this app."""
        paths = []
        sources = self.sources.copy() if self.sources else []
        if self.test_mode and self.test_sources:
            sources.extend(self.test_sources)

        for source in sources:
            path = "/".join(source.rsplit("/", 1)[:-1])
            if path not in paths:
                paths.append(path)
        return paths

    def all_sources(self) -> list[str]:
        """Get all sources of the application that should be copied to the app.

        :returns: The Path to the dist-info folder.
        """
        sources = self.sources.copy() if self.sources else []
        if self.test_mode and self.test_sources:
            sources.extend(self.test_sources)
        return sources

    def main_module(self):
        """The path to the main module for the app.

        In normal operation, this is ``app.module_name``; however,
        in test mode, it is prefixed with ``tests.``.
        """
        if self.test_mode:
            return f"tests.{self.module_name}"
        else:
            return self.module_name


def merge_config(config, data):
    """Merge a new set of configuration requirements into a base configuration.

    :param config: the base configuration to update. This configuration is modified in-
        situ.
    :param data: The new configuration data to merge into the configuration.
    """
    # Properties that are cumulative lists
    for option in [
        "requires",
        "sources",
        "test_requires",
        "test_sources",
    ]:
        value = data.pop(option, [])

        if value:
            config.setdefault(option, []).extend(value)

    # Properties that are cumulative tables
    for option in ["permission"]:
        value = data.pop(option, {})

        if value:
            config.setdefault(option, {}).update(value)

    config.update(data)


def merge_pep621_config(global_config, pep621_config):
    """Merge a PEP621 configuration into a Briefcase configuration."""

    if requires_python := pep621_config.get("requires-python"):
        global_config["requires_python"] = requires_python

    def maybe_update(field, *project_fields):
        # If there's an existing key in the Briefcase config, it takes priority.
        if field in global_config:
            return

        # Traverse the fields in the pep621 config; if the config exists, set it
        # in the Briefcase config.
        datum = pep621_config
        try:
            for key in project_fields:
                datum = datum[key]
        except KeyError:
            pass
        else:
            global_config[field] = datum

    # Keys that map directly
    maybe_update("description", "description")
    maybe_update("license", "license")
    maybe_update("url", "urls", "Homepage")
    maybe_update("version", "version")

    # Use the details of the first author as the Briefcase author.
    if "author" not in global_config:
        try:
            global_config["author"] = pep621_config["authors"][0]["name"]
        except (KeyError, IndexError):
            pass
    if "author_email" not in global_config:
        try:
            global_config["author_email"] = pep621_config["authors"][0]["email"]
        except (KeyError, IndexError):
            pass

    # Briefcase requires is cumulative over PEP621 dependencies
    try:
        pep621_dependencies = pep621_config["dependencies"]
        requires = global_config.get("requires", [])
        global_config["requires"] = pep621_dependencies + requires
    except KeyError:
        pass

    # Briefcase test_requires is cumulative over PEP621 optional test dependencies
    try:
        pep621_test_dependencies = pep621_config["optional-dependencies"]["test"]
        test_requires = global_config.get("test_requires", [])
        global_config["test_requires"] = pep621_test_dependencies + test_requires
    except KeyError:
        pass


def parse_config(config_file, platform, output_format, console):
    """Parse the briefcase section of the pyproject.toml configuration file.

    This method only does basic structural parsing of the TOML, looking for,
    at a minimum, a ``[tool.briefcase.app.<appname>]`` section declaring the
    existence of a single app. It will also search for:

      * ``[tool.briefcase]`` - global briefcase settings
      * ``[tool.briefcase.app.<appname>]`` - settings specific to the app
      * ``[tool.briefcase.app.<appname>.<platform>]`` - settings specific to
        the platform
      * ``[tool.briefcase.app.<appname>.<platform>.<format>]`` - settings
        specific to the output format

    A configuration can define multiple apps; the final output is the merged
    content of the global, app, platform and output format settings
    for each app, with output format definitions taking precedence over
    platform, over app-level, over global. The final result is a single
    (mostly) flat dictionary for each app.

    :param config_file: A file-like object containing TOML to be parsed.
    :param platform: The platform being targeted
    :param output_format: The output format
    :param console: The console to use for any output or logging.
    :returns: A dictionary of configuration data. The top level dictionary is
        keyed by the names of the apps that are declared; each value is
        itself the configuration data merged from global, app, platform and
        format definitions.
    """
    try:
        pyproject = tomllib.load(config_file)
    except tomllib.TOMLDecodeError as e:
        raise BriefcaseConfigError(f"Invalid pyproject.toml: {e}") from e

    try:
        global_config = pyproject["tool"]["briefcase"]
    except KeyError as e:
        raise BriefcaseConfigError("No tool.briefcase section in pyproject.toml") from e

    # Merge the PEP621 configuration (if it exists)
    try:
        merge_pep621_config(global_config, pyproject["project"])
    except KeyError:
        pass

    # For consistent results, sort the platforms and formats
    all_platforms = sorted(get_platforms().keys())
    all_formats = sorted(get_output_formats(platform).keys())

    try:
        all_apps = global_config.pop("app")
    except KeyError as e:
        raise BriefcaseConfigError("No Briefcase apps defined in pyproject.toml") from e

    for name, config in [("project", global_config), *all_apps.items()]:
        if isinstance(config.get("license"), str):
            section_name = "the Project" if name == "project" else f"{name!r}"
            console.warning(
                f"""
*************************************************************************
** {f"WARNING: License Definition for {section_name} is Deprecated":67} **
*************************************************************************

    Briefcase now uses PEP 621 format for license definitions.

    Previously, the name of the license was assigned to the 'license'
    field in pyproject.toml. For PEP 621, the name of the license is
    assigned to 'license.text' or the name of the file containing the
    license is assigned to 'license.file'.

    The current configuration for {section_name} has a 'license' field
    that is specified as a string:

        license = "{config["license"]}"

    To use the PEP 621 format (and to remove this warning), specify that
    the LICENSE file contains the license for {section_name}:

        license.file = "LICENSE"

*************************************************************************
"""
            )
            config["license"] = {"file": "LICENSE"}

    # Build the flat configuration for each app,
    # based on the requested platform and output format
    app_configs = {}
    for app_name, app_data in all_apps.items():
        # At this point, the base configuration will contain a section
        # for each configured platform. Iterate over all the known platforms,
        # and remove these platform configurations. Keep a copy of the platform
        # configuration if it matches the requested platform, and merge it
        # into the app's configuration
        platform_data = None
        for p in all_platforms:
            try:
                platform_block = app_data.pop(p)

                if p == platform:
                    # If the platform matches the requested format, preserve
                    # it for later use.
                    platform_data = platform_block
                    merge_config(platform_data, platform_data)

                    # The platform configuration will contain a section
                    # for each configured output format. Iterate over all
                    # the output format of the platform, and remove these
                    # output format configurations. Keep a copy of the output
                    # format configuration if it matches the requested output
                    # format, and merge it into the platform's configuration.
                    format_data = None
                    for f in all_formats:
                        try:
                            format_block = platform_data.pop(f)

                            if f == output_format:
                                # If the output format matches the requested
                                # one, preserve it.
                                format_data = format_block

                        except KeyError:
                            pass

                    # If we found a specific configuration for the requested
                    # output format, add it to the platform configuration,
                    # overwriting any platform-level settings with format-level
                    # values.
                    if format_data:
                        merge_config(platform_data, format_data)

            except KeyError:
                pass

        # Now construct the final configuration.
        # First, convert the requirement definition at the global level
        merge_config(global_config, global_config)

        # The app's config starts as a copy of the base briefcase configuration.
        config = copy.deepcopy(global_config)

        # The app name is both the key, and a property of the configuration
        config["app_name"] = app_name

        # Merge the app-specific requirements
        merge_config(config, app_data)

        # If there is platform-specific configuration, merge the requirements,
        # then overwrite the platform-specific values.
        # This will already include any format-specific configuration.
        if platform_data:
            merge_config(config, platform_data)

        # Construct a configuration object, and add it to the list
        # of configurations that are being handled.
        app_configs[app_name] = config

    return global_config, app_configs



from briefcase.debuggers.base import (
    BaseDebugger,
    DebuggerConnectionMode,
    get_debugger_requirement,
)


class PdbDebugger(BaseDebugger):
    """Definition for a plugin that defines a new Briefcase debugger."""

    @property
    def name(self) -> str:
        """Return the name debugger."""
        return "pdb"

    @property
    def connection_mode(self) -> DebuggerConnectionMode:
        """Return the connection mode of the debugger."""
        return DebuggerConnectionMode.SERVER

    @property
    def debugger_support_pkg(self) -> str:
        """Get the name of the debugger support package."""
        return get_debugger_requirement("briefcase-debugger", "[pdb]")



from unittest.mock import MagicMock

import pytest

from briefcase.console import Console
from briefcase.integrations.virtual_environment import NoOpVenvContext, VenvContext


@pytest.fixture
def dummy_console():
    return MagicMock(spec_set=Console)


@pytest.fixture
def venv_path(tmp_path):
    return tmp_path / "test_venv"


@pytest.fixture
def venv_context(mock_tools, venv_path):
    return VenvContext(
        tools=mock_tools,
        venv_path=venv_path,
    )


@pytest.fixture
def noop_context(mock_tools, venv_path):
    return NoOpVenvContext(tools=mock_tools, venv_path=venv_path)


@pytest.fixture
def mock_subprocess_setup(venv_context, monkeypatch):
    """Setup mock objects for subprocess testing across all VenvContext tests."""
    mock_rewrite_head = MagicMock(return_value=["rewritten", "args"])
    mock_full_env = MagicMock(return_value={"FULL": "env"})
    mock_subprocess = MagicMock()

    mock_popen_instance = MagicMock()
    mock_completed_process = MagicMock()

    mock_subprocess.Popen.return_value = mock_popen_instance
    mock_subprocess.run.return_value = mock_completed_process
    mock_subprocess.check_output.return_value = "output"

    monkeypatch.setattr(venv_context, "_rewrite_head", mock_rewrite_head)
    monkeypatch.setattr(venv_context, "full_env", mock_full_env)
    monkeypatch.setattr(venv_context.tools, "subprocess", mock_subprocess)

    return {
        "rewrite_head": mock_rewrite_head,
        "full_env": mock_full_env,
        "subprocess": mock_subprocess,
        "popen_instance": mock_popen_instance,
        "completed_process": mock_completed_process,
    }


@pytest.fixture
def mock_noop_subprocess_setup(noop_context, monkeypatch):
    """Setup mock objects for subprocess testing across all NoOpVenvContext tests."""
    mock_subprocess = MagicMock()

    mock_popen_instance = MagicMock()
    mock_completed_process = MagicMock()

    mock_subprocess.Popen.return_value = mock_popen_instance
    mock_subprocess.run.return_value = mock_completed_process
    mock_subprocess.check_output.return_value = "output"

    monkeypatch.setattr(noop_context.tools, "subprocess", mock_subprocess)

    return {
        "subprocess": mock_subprocess,
        "popen_instance": mock_popen_instance,
        "completed_process": mock_completed_process,
    }



import pytest

from briefcase.platforms.windows.visualstudio import WindowsVisualStudioCreateCommand

# Most tests and fixtures are the same for both "app" and "visualstudio". This file only
# contains those that need to be overridden.
from ..app.create.test_create import *  # noqa: F403


@pytest.fixture
def create_command(dummy_console, tmp_path):
    return WindowsVisualStudioCreateCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


def test_package_path(create_command, first_app_config, tmp_path):
    """The default package_path is passed as an absolute path."""
    context = create_command.output_format_template_context(first_app_config)
    assert context["package_path"] == str(
        tmp_path / "base_path/build/first-app/windows/visualstudio/x64/Release"
    )



import pytest

from tests.integrations.virtual_environment.utils import (
    ENVIRONMENT_TEST_PARAMS,
    assert_environment_handling,
)


@pytest.mark.parametrize(("env_override", "other_kwargs"), ENVIRONMENT_TEST_PARAMS)
def test_run_environment_handling(
    venv_context,
    mock_subprocess_setup,
    env_override,
    other_kwargs,
):
    """Run merges venv with overrides and passes all kwargs to subprocess."""
    mocks = mock_subprocess_setup

    kwargs = other_kwargs.copy()
    if env_override is not None:
        kwargs["env"] = env_override

    original_kwargs = kwargs.copy()

    result = venv_context.run(["test"], **kwargs)

    assert kwargs == original_kwargs
    mocks["rewrite_head"].assert_called_once_with(["test"])

    # The venv's full_env should be called with any provided env override, and the
    # resulting environment should be passed to subprocess.check_output along with other
    # kwargs unchanged.
    assert_environment_handling(
        mock_full_env=mocks["full_env"],
        env_override=env_override,
        mock_method=mocks["subprocess"].run,
        method_args=["rewritten", "args"],
        other_kwargs=other_kwargs,
    )

    assert result is mocks["subprocess"].run.return_value


def test_run_kwargs_env_extraction(venv_context, mock_subprocess_setup):
    """Run extracts env from kwargs without modifying the original kwargs."""
    mocks = mock_subprocess_setup

    original_kwargs = {
        "env": {"CUSTOM": "value"},
        "cwd": "/tmp",
        "check": True,
        "capture_output": False,
    }
    kwargs_copy = original_kwargs.copy()

    result = venv_context.run(["test"], **original_kwargs)

    # The original kwargs dict should remain unchanged
    assert original_kwargs == kwargs_copy

    # The env value is passed to full_env()
    mocks["full_env"].assert_called_once_with({"CUSTOM": "value"})

    # The resulting merged environment replaces the original env in the subprocess call.
    expected_call_kwargs = {
        "cwd": "/tmp",
        "check": True,
        "capture_output": False,
        "env": {"FULL": "env"},
    }
    mocks["subprocess"].run.assert_called_once_with(
        ["rewritten", "args"], **expected_call_kwargs
    )

    assert result is mocks["subprocess"].run.return_value



from unittest.mock import MagicMock

import pytest

from briefcase.integrations.cookiecutter import GradleEscape


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("Hello World", "Hello World"),
        ("Hello ' World", "Hello \\' World"),
        ("Hello \\ World", "Hello \\\\ World"),
    ],
)
def test_escape_gradle(value, expected):
    env = MagicMock()
    env.filters = {}
    GradleEscape(env)
    assert env.filters["escape_gradle"](value) == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("helloworld", "helloworld"),
        ("helloworldı", '"helloworldı"'),  # noqa: RUF001 (ambiguous non-ASCII symbol)
    ],
)
def test_escape_non_ascii(value, expected):
    env = MagicMock()
    env.filters = {}
    GradleEscape(env)
    assert env.filters["escape_non_ascii"](value) == expected



import json
import os
from unittest.mock import MagicMock

import briefcase_debugger
import briefcase_debugger.pdb
import pytest


def test_no_env_vars(monkeypatch, capsys):
    """Nothing happens, when no env vars are set."""
    os_environ = {}
    monkeypatch.setattr(os, "environ", os_environ)

    # start test function
    briefcase_debugger.start_remote_debugger()

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == ""


def test_no_debugger_verbose(monkeypatch, capsys):
    """Nothing happens except a short message, when only verbose is requested."""
    os_environ = {}
    os_environ["BRIEFCASE_DEBUG"] = "1"
    monkeypatch.setattr(os, "environ", os_environ)

    # start test function
    briefcase_debugger.start_remote_debugger()

    captured = capsys.readouterr()
    assert (
        captured.out
        == "No 'BRIEFCASE_DEBUGGER' environment variable found. Debugger not starting.\n"
    )
    assert captured.err == ""


@pytest.mark.parametrize("verbose", [True, False])
@pytest.mark.parametrize(
    ("host_os", "expected_host_cmds"),
    [
        ("Windows", ["telnet somehost 9999"]),
        ("Darwin", ["nc somehost 9999"]),
        ("Linux", ["nc somehost 9999"]),
        ("UnknownOS", ["nc somehost 9999", "telnet somehost 9999"]),
    ],
)
def test_with_debugger(monkeypatch, host_os, expected_host_cmds, capsys, verbose):
    """Normal debug session."""
    os_environ = {}
    os_environ["BRIEFCASE_DEBUG"] = "1" if verbose else "0"
    os_environ["BRIEFCASE_DEBUGGER"] = json.dumps(
        {
            "debugger": "pdb",
            "host": "somehost",
            "port": 9999,
            "host_os": host_os,
        }
    )
    monkeypatch.setattr(os, "environ", os_environ)

    fake_remote_pdb = MagicMock()
    monkeypatch.setattr(briefcase_debugger.pdb, "RemotePdb", fake_remote_pdb)

    # start test function
    briefcase_debugger.start_remote_debugger()

    fake_remote_pdb.assert_called_once_with(
        "somehost",
        9999,
        quiet=True,
    )

    captured = capsys.readouterr()
    assert "Waiting for debugger to attach..." in captured.out
    for cmd in expected_host_cmds:
        assert cmd in captured.out
    assert captured.err == ""



import shutil
import subprocess
import sys
from pathlib import Path
from unittest import mock

import pytest

from briefcase.exceptions import BriefcaseCommandError
from briefcase.platforms.linux import system
from briefcase.platforms.linux.system import (
    LinuxSystemPackageCommand,
    debian_multiline_description,
)

from ....utils import create_file


@pytest.fixture
def package_command(dummy_console, first_app, tmp_path):
    command = LinuxSystemPackageCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.home_path = tmp_path / "home"

    # Mock ABI from packaging system
    command._deb_abi = "wonky"

    # Mock the app context
    command.tools.app_tools[first_app].app_context = mock.MagicMock()

    # Mock shutil
    command.tools.shutil = mock.MagicMock()
    # Make the mock rmtree still remove content
    command.tools.shutil.rmtree.side_effect = shutil.rmtree

    # Mock not using docker
    command.target_image = None
    command.extra_docker_build_args = []

    return command


@pytest.fixture
def first_app_deb(first_app):
    # Mock a debian app
    first_app.python_version_tag = "3.10"
    first_app.target_vendor_base = "debian"
    first_app.packaging_format = "deb"
    first_app.glibc_version = "2.99"
    first_app.long_description = "Long description\nfor the app"

    return first_app


@pytest.fixture
def underscore_app_deb(underscore_app):
    # Mock a debian app
    underscore_app.python_version_tag = "3.10"
    underscore_app.target_vendor_base = "debian"
    underscore_app.packaging_format = "deb"
    underscore_app.glibc_version = "2.99"
    underscore_app.long_description = "Long description\nfor the app"

    return underscore_app


@pytest.fixture
def external_first_app_deb(first_app_deb, tmp_path):
    # Make the app external
    first_app_deb.sources = None
    first_app_deb.external_package_path = tmp_path / "base_path/external/package-deb"

    # Move the generated first app to the external location
    (tmp_path / "base_path/external").mkdir()
    shutil.move(
        tmp_path / "base_path/build/first-app/somevendor/surprising/first-app-0.0.1",
        tmp_path / "base_path/external/package-deb",
    )
    return first_app_deb


def test_verify_no_docker(package_command, first_app_deb, monkeypatch):
    """If not using docker, existence of dpkg-deb is verified."""
    # Mock the existence of dpkg-deb
    package_command.tools.shutil.which = mock.MagicMock(return_value="/mybin/dpkg-deb")

    # App tools can be verified
    package_command.verify_app_tools(first_app_deb)

    # dpkg_deb was inspected
    package_command.tools.shutil.which.assert_called_once_with("dpkg-deb")


@pytest.mark.parametrize(
    ("vendor_base", "error_msg"),
    [
        (
            "debian",
            "Can't find the dpkg tools. Try running `sudo apt install dpkg-dev`.",
        ),
        (None, "Can't find the dpkg-deb tool. Install this first to package the deb."),
    ],
)
def test_verify_dpkg_deb_missing(
    package_command,
    first_app_deb,
    vendor_base,
    error_msg,
    monkeypatch,
):
    """If dpkg-deb isn't installed, an error is raised."""
    # Mock distro so packager is found or not appropriately
    first_app_deb.target_vendor_base = vendor_base

    # Mock packager as missing
    package_command.tools.shutil.which = mock.MagicMock(return_value="")

    # Verifying app tools will raise an error
    with pytest.raises(BriefcaseCommandError, match=error_msg):
        package_command.verify_app_tools(first_app_deb)

    # which was called for dpkg-deb
    package_command.tools.shutil.which.assert_called_once_with("dpkg-deb")


def test_verify_docker(package_command, first_app_deb, monkeypatch):
    """If using Docker, no tool checks are needed."""
    # Mock using docker
    package_command.target_image = "somevendor:surprising"

    # Mock the existence of a valid non-docker system Python
    # with the same major/minor as the current Python,
    # plus the existence of dpkg-deb
    python3 = mock.MagicMock()
    python3.resolve.return_value = Path(
        f"/usr/bin/python{sys.version_info.major}.{sys.version_info.minor}"
    )

    dpkg_deb = mock.MagicMock()
    dpkg_deb.exists.return_value = False

    mock_Path = mock.MagicMock(side_effect=[python3, dpkg_deb])
    monkeypatch.setattr(system, "Path", mock_Path)

    # App tools can be verified
    package_command.verify_app_tools(first_app_deb)

    # dpkg_deb was not inspected
    dpkg_deb.exists.assert_not_called()


@pytest.mark.skipif(sys.platform == "win32", reason="Can't build debs on Windows")
def test_deb_package(package_command, first_app_deb, tmp_path):
    """A deb app can be packaged."""
    bundle_path = tmp_path / "base_path/build/first-app/somevendor/surprising"

    # Package the app
    package_command.package_app(first_app_deb)

    # The control file is written
    assert (bundle_path / "first-app-0.0.1/DEBIAN/control").exists()
    with (bundle_path / "first-app-0.0.1/DEBIAN/control").open(encoding="utf-8") as f:
        assert (
            f.read()
            == "\n".join(
                [
                    "Package: first-app",
                    "Version: 0.0.1",
                    "Architecture: wonky",
                    "Maintainer: Megacorp <maintainer@example.com>",
                    "Homepage: https://example.com/first-app",
                    "Description: The first simple app \\ demonstration",
                    " Long description",
                    " for the app",
                    "Depends: libc6 (>=2.99), libpython3.10",
                    "Section: utils",
                    "Priority: optional",
                ]
            )
            + "\n"
        )

    package_command.tools.app_tools[
        first_app_deb
    ].app_context.run.assert_called_once_with(
        [
            "dpkg-deb",
            "--build",
            "--root-owner-group",
            bundle_path / "first-app-0.0.1",
        ],
        check=True,
        cwd=bundle_path,
    )

    # The deb was moved into the final location
    package_command.tools.shutil.move.assert_called_once_with(
        bundle_path / "first-app-0.0.1.deb",
        tmp_path / "base_path/dist/first-app_0.0.1-1~somevendor-surprising_wonky.deb",
    )


def test_deb_re_package(package_command, first_app_deb, tmp_path):
    """A deb app that has previously been packaged can be re-packaged."""
    bundle_path = tmp_path / "base_path/build/first-app/somevendor/surprising"

    # Create an old control file that will be overwritten.
    create_file(bundle_path / "first-app-0.0.1/DEBIAN/control", "Old control content")

    # Package the app
    package_command.package_app(first_app_deb)

    # The control file is re-written
    assert (bundle_path / "first-app-0.0.1/DEBIAN/control").exists()
    with (bundle_path / "first-app-0.0.1/DEBIAN/control").open(encoding="utf-8") as f:
        assert (
            f.read()
            == "\n".join(
                [
                    "Package: first-app",
                    "Version: 0.0.1",
                    "Architecture: wonky",
                    "Maintainer: Megacorp <maintainer@example.com>",
                    "Homepage: https://example.com/first-app",
                    "Description: The first simple app \\ demonstration",
                    " Long description",
                    " for the app",
                    "Depends: libc6 (>=2.99), libpython3.10",
                    "Section: utils",
                    "Priority: optional",
                ]
            )
            + "\n"
        )

    package_command.tools.app_tools[
        first_app_deb
    ].app_context.run.assert_called_once_with(
        [
            "dpkg-deb",
            "--build",
            "--root-owner-group",
            bundle_path / "first-app-0.0.1",
        ],
        check=True,
        cwd=bundle_path,
    )

    # The deb was moved into the final location
    package_command.tools.shutil.move.assert_called_once_with(
        bundle_path / "first-app-0.0.1.deb",
        tmp_path / "base_path/dist/first-app_0.0.1-1~somevendor-surprising_wonky.deb",
    )


@pytest.mark.skipif(sys.platform == "win32", reason="Can't build debs on Windows")
def test_deb_package_underscore(package_command, underscore_app_deb, tmp_path):
    """A deb app can be packaged."""
    package_command.tools.app_tools[underscore_app_deb].app_context = mock.MagicMock()

    bundle_path = tmp_path / "base_path/build/underscore_app/somevendor/surprising"

    # Package the app
    package_command.package_app(underscore_app_deb)

    # The control file is written
    assert (bundle_path / "underscore_app-0.0.1/DEBIAN/control").exists()
    with (bundle_path / "underscore_app-0.0.1/DEBIAN/control").open(
        encoding="utf-8"
    ) as f:
        assert (
            f.read()
            == "\n".join(
                [
                    "Package: underscore-app",
                    "Version: 0.0.1",
                    "Architecture: wonky",
                    "Maintainer: Megacorp <maintainer@example.com>",
                    "Homepage: https://example.com/underscore_app",
                    "Description: The first simple app \\ demonstration",
                    " Long description",
                    " for the app",
                    "Depends: libc6 (>=2.99), libpython3.10",
                    "Section: utils",
                    "Priority: optional",
                ]
            )
            + "\n"
        )

    package_command.tools.app_tools[
        underscore_app_deb
    ].app_context.run.assert_called_once_with(
        [
            "dpkg-deb",
            "--build",
            "--root-owner-group",
            bundle_path / "underscore_app-0.0.1",
        ],
        check=True,
        cwd=bundle_path,
    )

    # The deb was moved into the final location
    package_command.tools.shutil.move.assert_called_once_with(
        bundle_path / "underscore_app-0.0.1.deb",
        tmp_path
        / "base_path/dist/underscore-app_0.0.1-1~somevendor-surprising_wonky.deb",
    )


def test_deb_package_no_long_description(package_command, first_app_deb, tmp_path):
    """A deb app without a long description raises an error."""
    bundle_path = tmp_path / "base_path/build/first-app/somevendor/surprising"

    # Delete the long description
    first_app_deb.long_description = None

    # Packaging the app will fail
    with pytest.raises(
        BriefcaseCommandError,
        match=r"App configuration does not define `long_description`. Debian projects require a long description.",
    ):
        package_command.package_app(first_app_deb)

    # The control file won't be written
    assert not (bundle_path / "first-app-0.0.1/DEBIAN/control").exists()


@pytest.mark.parametrize(
    ("input", "output"),
    [
        ("", ""),
        ("one line", "one line"),
        ("first line\nsecond line", "first line\n second line"),
        ("first line\n\nsecond line", "first line\n second line"),
        ("first line\n  \nsecond line", "first line\n second line"),
    ],
)
def test_multiline_long_description(input, output):
    """Multiline debian descriptions are transformed."""
    assert debian_multiline_description(input) == output


def test_deb_package_extra_requirements(package_command, first_app_deb, tmp_path):
    """A deb app can be packaged with extra runtime requirements and configuration
    options."""
    bundle_path = tmp_path / "base_path/build/first-app/somevendor/surprising"

    # Add system requirements and other optional settings.
    first_app_deb.system_runtime_requires = ["first", "second (>=1.2.3)"]
    first_app_deb.revision = 42
    first_app_deb.system_section = "Funny stuff"

    # Package the app
    package_command.package_app(first_app_deb)

    # The control file is written
    assert (bundle_path / "first-app-0.0.1/DEBIAN/control").exists()
    with (bundle_path / "first-app-0.0.1/DEBIAN/control").open(encoding="utf-8") as f:
        assert (
            f.read()
            == "\n".join(
                [
                    "Package: first-app",
                    "Version: 0.0.1",
                    "Architecture: wonky",
                    "Maintainer: Megacorp <maintainer@example.com>",
                    "Homepage: https://example.com/first-app",
                    "Description: The first simple app \\ demonstration",
                    " Long description",
                    " for the app",
                    "Depends: libc6 (>=2.99), libpython3.10, first, second (>=1.2.3)",
                    "Section: Funny stuff",
                    "Priority: optional",
                ]
            )
            + "\n"
        )

    package_command.tools.app_tools[
        first_app_deb
    ].app_context.run.assert_called_once_with(
        [
            "dpkg-deb",
            "--build",
            "--root-owner-group",
            bundle_path / "first-app-0.0.1",
        ],
        check=True,
        cwd=bundle_path,
    )

    # The deb was moved into the final location
    package_command.tools.shutil.move.assert_called_once_with(
        bundle_path / "first-app-0.0.1.deb",
        tmp_path / "base_path/dist/first-app_0.0.1-42~somevendor-surprising_wonky.deb",
    )


def test_deb_package_failure(package_command, first_app_deb, tmp_path):
    """If a packaging doesn't succeed, an error is raised."""
    bundle_path = tmp_path / "base_path/build/first-app/somevendor/surprising"

    # Mock a packaging failure
    package_command.tools.app_tools[
        first_app_deb
    ].app_context.run.side_effect = subprocess.CalledProcessError(
        cmd="dpkg-deb ...", returncode=-1
    )

    # Package the app
    with pytest.raises(
        BriefcaseCommandError, match=r"Error while building .deb package for first-app."
    ):
        package_command.package_app(first_app_deb)

    # The control file is written
    assert (bundle_path / "first-app-0.0.1/DEBIAN/control").exists()
    with (bundle_path / "first-app-0.0.1/DEBIAN/control").open(encoding="utf-8") as f:
        assert (
            f.read()
            == "\n".join(
                [
                    "Package: first-app",
                    "Version: 0.0.1",
                    "Architecture: wonky",
                    "Maintainer: Megacorp <maintainer@example.com>",
                    "Homepage: https://example.com/first-app",
                    "Description: The first simple app \\ demonstration",
                    " Long description",
                    " for the app",
                    "Depends: libc6 (>=2.99), libpython3.10",
                    "Section: utils",
                    "Priority: optional",
                ]
            )
            + "\n"
        )

    # The call to package was made
    package_command.tools.app_tools[
        first_app_deb
    ].app_context.run.assert_called_once_with(
        [
            "dpkg-deb",
            "--build",
            "--root-owner-group",
            bundle_path / "first-app-0.0.1",
        ],
        check=True,
        cwd=bundle_path,
    )

    # The deb wasn't built, so it wasn't moved.
    package_command.tools.shutil.move.assert_not_called()


@pytest.mark.skipif(sys.platform == "win32", reason="Can't build debs on Windows")
def test_external_deb_package_docker(package_command, external_first_app_deb, tmp_path):
    """An external app cannot be packaged as a deb using Docker."""
    bundle_path = tmp_path / "base_path/build/first-app/somevendor/surprising"
    package_path = tmp_path / "base_path/external/package-deb"

    # Package the app
    package_command.package_app(external_first_app_deb)

    # The control file is written
    assert (package_path / "DEBIAN/control").exists()
    with (package_path / "DEBIAN/control").open(encoding="utf-8") as f:
        assert (
            f.read()
            == "\n".join(
                [
                    "Package: first-app",
                    "Version: 0.0.1",
                    "Architecture: wonky",
                    "Maintainer: Megacorp <maintainer@example.com>",
                    "Homepage: https://example.com/first-app",
                    "Description: The first simple app \\ demonstration",
                    " Long description",
                    " for the app",
                    "Depends: libc6 (>=2.99), libpython3.10",
                    "Section: utils",
                    "Priority: optional",
                ]
            )
            + "\n"
        )

    package_command.tools.app_tools[
        external_first_app_deb
    ].app_context.run.assert_called_once_with(
        [
            "dpkg-deb",
            "--build",
            "--root-owner-group",
            package_path,
        ],
        check=True,
        cwd=bundle_path,
    )

    # The deb was moved into the final location
    package_command.tools.shutil.move.assert_called_once_with(
        tmp_path / "base_path/external/package-deb.deb",
        tmp_path / "base_path/dist/first-app_0.0.1-1~somevendor-surprising_wonky.deb",
    )



import os
import sys

import briefcase_debugger.debugpy
import pytest
from briefcase_debugger.config import (
    AppPackagesPathMappings,
    AppPathMappings,
    DebuggerConfig,
)


def test_mappings_not_existing():
    """Complete empty config."""
    path_mappings = briefcase_debugger.debugpy.load_path_mappings({}, False)
    assert path_mappings == []


def test_mappings_none(monkeypatch):
    """Config with no mappings set."""
    config = DebuggerConfig(
        debugger="debugpy",
        host="",
        port=0,
        app_path_mappings=None,
        app_packages_path_mappings=None,
    )
    path_mappings = briefcase_debugger.debugpy.load_path_mappings(config, False)
    assert path_mappings == []


@pytest.mark.parametrize(
    (
        "os_name",
        "app_path_mappings",
        "app_packages_path_mappings",
        "sys_path",
        "expected_path_mappings",
    ),
    [
        # Windows
        pytest.param(
            "nt",
            AppPathMappings(
                device_sys_path_regex="app$",
                device_subfolders=["helloworld"],
                host_folders=["C:\\PROJECT_ROOT\\src\\helloworld"],
            ),
            None,
            [
                "C:\\PROJECT_ROOT\\build\\helloworld\\windows\\app\\src\\python313.zip",
                "C:\\PROJECT_ROOT\\build\\helloworld\\windows\\app\\src",
                "C:\\PROJECT_ROOT\\build\\helloworld\\windows\\app\\src\\app",
                "C:\\PROJECT_ROOT\\build\\helloworld\\windows\\app\\src\\app_packages",
            ],
            [
                (
                    "C:\\PROJECT_ROOT\\src\\helloworld",
                    "C:\\PROJECT_ROOT\\build\\helloworld\\windows\\app\\src\\app\\helloworld",
                ),
            ],
            id="windows",
        ),
        # Windows with `app_packages_path_mappings` (currently not used by briefcase, but principally possible)
        pytest.param(
            "nt",
            AppPathMappings(
                device_sys_path_regex="app$",
                device_subfolders=["helloworld"],
                host_folders=["C:\\PROJECT_ROOT\\src\\helloworld"],
            ),
            AppPackagesPathMappings(
                sys_path_regex="app_packages$",
                host_folder="C:\\PROJECT_ROOT\\build\\helloworld\\windows\\app\\src\\app_packages",
            ),
            [
                "C:\\PROJECT_ROOT\\build\\helloworld\\windows\\app\\src\\python313.zip",
                "C:\\PROJECT_ROOT\\build\\helloworld\\windows\\app\\src",
                "C:\\PROJECT_ROOT\\build\\helloworld\\windows\\app\\src\\app",
                "C:\\PROJECT_ROOT\\build\\helloworld\\windows\\app\\src\\app_packages",
            ],
            [
                (
                    "C:\\PROJECT_ROOT\\src\\helloworld",
                    "C:\\PROJECT_ROOT\\build\\helloworld\\windows\\app\\src\\app\\helloworld",
                ),
                (
                    "C:\\PROJECT_ROOT\\build\\helloworld\\windows\\app\\src\\app_packages",
                    "C:\\PROJECT_ROOT\\build\\helloworld\\windows\\app\\src\\app_packages",
                ),
            ],
            id="windows-with-app-packages",
        ),
        # macOS
        pytest.param(
            "posix",
            AppPathMappings(
                device_sys_path_regex="app$",
                device_subfolders=["helloworld"],
                host_folders=["/PROJECT_ROOT/src/helloworld"],
            ),
            None,
            [
                "/PROJECT_ROOT/build/helloworld/macos/app/Hello World.app/Contents/Frameworks/Python.framework/Versions/3.13/lib/python3.13",
                "/PROJECT_ROOT/build/helloworld/macos/app/Hello World.app/Contents/Frameworks/Python.framework/Versions/3.13/lib/python3.13/lib-dynload",
                "/PROJECT_ROOT/build/helloworld/macos/app/Hello World.app/Contents/Resources/app",
                "/PROJECT_ROOT/build/helloworld/macos/app/Hello World.app/Contents/Frameworks/Python.framework/Versions/3.13/lib/python3.13/site-packages",
                "/PROJECT_ROOT/build/helloworld/macos/app/Hello World.app/Contents/Resources/app_packages",
            ],
            [
                (
                    "/PROJECT_ROOT/src/helloworld",
                    "/PROJECT_ROOT/build/helloworld/macos/app/Hello World.app/Contents/Resources/app/helloworld",
                )
            ],
            id="macos",
        ),
        # iOS
        pytest.param(
            "posix",
            AppPathMappings(
                device_sys_path_regex="app$",
                device_subfolders=["helloworld"],
                host_folders=["/PROJECT_ROOT/src/helloworld"],
            ),
            AppPackagesPathMappings(
                sys_path_regex="app_packages$",
                host_folder="/APP_PACKAGES_PATH/app_packages.iphonesimulator",
            ),
            [
                "CoreSimulator/Devices/RANDOM_NUMBER/data/Containers/Bundle/Application/RANDOM_NUMBER/Hello World.app/python/lib/python3.13",
                "CoreSimulator/Devices/RANDOM_NUMBER/data/Containers/Bundle/Application/RANDOM_NUMBER/Hello World.app/python/lib/python3.13/lib-dynload",
                "CoreSimulator/Devices/RANDOM_NUMBER/data/Containers/Bundle/Application/RANDOM_NUMBER/Hello World.app/app",
                "CoreSimulator/Devices/RANDOM_NUMBER/data/Containers/Bundle/Application/RANDOM_NUMBER/Hello World.app/python/lib/python3.13/site-packages",
                "CoreSimulator/Devices/RANDOM_NUMBER/data/Containers/Bundle/Application/RANDOM_NUMBER/Hello World.app/app_packages",
            ],
            [
                (
                    "/PROJECT_ROOT/src/helloworld",
                    "CoreSimulator/Devices/RANDOM_NUMBER/data/Containers/Bundle/Application/RANDOM_NUMBER/Hello World.app/app/helloworld",
                ),
                (
                    "/APP_PACKAGES_PATH/app_packages.iphonesimulator",
                    "CoreSimulator/Devices/RANDOM_NUMBER/data/Containers/Bundle/Application/RANDOM_NUMBER/Hello World.app/app_packages",
                ),
            ],
            id="ios",
        ),
        # Android (with VS Code running on Windows)
        pytest.param(
            "posix",
            AppPathMappings(
                device_sys_path_regex="app$",
                device_subfolders=["helloworld"],
                host_folders=["C:\\PROJECT_ROOT\\src\\helloworld"],
            ),
            AppPackagesPathMappings(
                sys_path_regex="requirements$",
                host_folder="C:\\BUNDLE_PATH\\app\\build\\python\\pip\\debug\\common",
            ),
            [
                "/data/data/com.example.helloworld/files/chaquopy/AssetFinder/app",
                "/data/data/com.example.helloworld/files/chaquopy/AssetFinder/requirements",
                "/data/data/com.example.helloworld/files/chaquopy/AssetFinder/stdlib-x86_64",
                "/data/user/0/com.example.helloworld/files/chaquopy/stdlib-common.imy",
                "/data/user/0/com.example.helloworld/files/chaquopy/bootstrap.imy",
                "/data/user/0/com.example.helloworld/files/chaquopy/bootstrap-native/x86_64",
            ],
            [
                (
                    "C:\\PROJECT_ROOT\\src\\helloworld",
                    "/data/data/com.example.helloworld/files/chaquopy/AssetFinder/app/helloworld",
                ),
                (
                    "C:\\BUNDLE_PATH\\app\\build\\python\\pip\\debug\\common",
                    "/data/data/com.example.helloworld/files/chaquopy/AssetFinder/requirements",
                ),
            ],
            id="android-on-windows-host",
        ),
        # Android (with VS Code running on POSIX system)
        pytest.param(
            "posix",
            AppPathMappings(
                device_sys_path_regex="app$",
                device_subfolders=["helloworld"],
                host_folders=["/PROJECT_ROOT/src/helloworld"],
            ),
            AppPackagesPathMappings(
                sys_path_regex="requirements$",
                host_folder="/BUNDLE_PATH/app/build/python/pip/debug/common",
            ),
            [
                "/data/data/com.example.helloworld/files/chaquopy/AssetFinder/app",
                "/data/data/com.example.helloworld/files/chaquopy/AssetFinder/requirements",
                "/data/data/com.example.helloworld/files/chaquopy/AssetFinder/stdlib-x86_64",
                "/data/user/0/com.example.helloworld/files/chaquopy/stdlib-common.imy",
                "/data/user/0/com.example.helloworld/files/chaquopy/bootstrap.imy",
                "/data/user/0/com.example.helloworld/files/chaquopy/bootstrap-native/x86_64",
            ],
            [
                (
                    "/PROJECT_ROOT/src/helloworld",
                    "/data/data/com.example.helloworld/files/chaquopy/AssetFinder/app/helloworld",
                ),
                (
                    "/BUNDLE_PATH/app/build/python/pip/debug/common",
                    "/data/data/com.example.helloworld/files/chaquopy/AssetFinder/requirements",
                ),
            ],
            id="android-on-posix-host",
        ),
    ],
)
def test_mappings(
    os_name: str,
    app_path_mappings: AppPathMappings,
    app_packages_path_mappings: AppPackagesPathMappings | None,
    sys_path: list[str],
    expected_path_mappings: list[tuple[str, str]],
    monkeypatch,
):
    if os.name != os_name:
        pytest.skip(f"Test only runs on {os_name} systems")

    config = DebuggerConfig(
        debugger="debugpy",
        host="",
        port=0,
        app_path_mappings=app_path_mappings,
        app_packages_path_mappings=app_packages_path_mappings,
    )

    monkeypatch.setattr(sys, "path", sys_path)

    path_mappings = briefcase_debugger.debugpy.load_path_mappings(config, False)

    assert path_mappings == expected_path_mappings


@pytest.mark.parametrize(
    ("os_name", "app_path_mappings"),
    [
        # Windows
        pytest.param(
            "nt",
            AppPathMappings(
                device_sys_path_regex="app$",
                device_subfolders=["helloworld"],
                host_folders=["C:\\PROJECT_ROOT\\src\\helloworld"],
            ),
            id="windows",
        ),
        # POSIX (macOS/iOS/Android)
        pytest.param(
            "posix",
            AppPathMappings(
                device_sys_path_regex="app$",
                device_subfolders=["helloworld"],
                host_folders=["/PROJECT_ROOT/src/helloworld"],
            ),
            id="posix",
        ),
    ],
)
def test_mappings_wrong_sys_path(
    os_name: str,
    app_path_mappings: AppPathMappings,
    monkeypatch,
):
    """Path mappings with a wrong sys path set."""
    if os.name != os_name:
        pytest.skip(f"Test only runs on {os_name} systems")

    config = DebuggerConfig(
        debugger="debugpy",
        host="",
        port=0,
        app_path_mappings=app_path_mappings,
        app_packages_path_mappings=None,
    )

    sys_path = []
    monkeypatch.setattr(sys, "path", sys_path)

    with pytest.raises(ValueError, match=r"No sys.path entry matches regex"):
        briefcase_debugger.debugpy.load_path_mappings(config, False)



import subprocess
import sys
from pathlib import Path
from unittest import mock

import pytest

from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.linux.system import LinuxSystemDevCommand

from .test_run import mock_linux_env


@pytest.fixture
def dev_command(monkeypatch, dummy_console, first_app, tmp_path):
    command = LinuxSystemDevCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
        apps={"app": first_app},
    )

    # Default to running on Linux
    command.tools.host_os = "Linux"

    # Set the host architecture for test purposes.
    command.tools.host_arch = "wonky"

    mock_linux_env(command, tmp_path, monkeypatch)

    (tmp_path / "base_path/src").mkdir(parents=True)

    command.tools.home_path = tmp_path / "home"
    command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)
    command._stream_app_logs = mock.MagicMock()
    return command


def test_dev_system_app_starts(dev_command, first_app, tmp_path):
    """A Linux system app can be started in development mode."""
    log_popen = mock.MagicMock()
    dev_command.tools.subprocess.Popen.return_value = log_popen

    dev_command.verify_system_packages = mock.MagicMock()

    # Parse the command line
    dev_command.parse_options([])

    # The command runs without error
    dev_command()

    # Running dev will create the dev environment; this will verify system packages.
    dev_command.verify_system_packages.assert_called_once_with(first_app)

    # The app was started with streamed logs
    popen_args, popen_kwargs = dev_command.tools.subprocess.Popen.call_args

    # Check Python executable isn't the system executable.
    # It should be a venv in the base path.
    assert popen_args[0][0] != sys.executable
    assert (
        Path(popen_args[0][0]).parent.parent.parent
        == dev_command.base_path / ".briefcase/first-app"
    )
    assert (
        Path(popen_args[0][0])
        .parts[-3]
        .startswith("dev.cp3" if sys.platform == "win32" else "dev.cpython-3")
    )
    assert Path(popen_args[0][0]).parts[-2] == (
        "Scripts" if sys.platform == "win32" else "bin"
    )

    # Check that module name is in the inline Python command
    assert "run_module" in popen_args[0][2]
    assert first_app.module_name in popen_args[0][2]

    assert popen_kwargs["cwd"] == tmp_path / "home"
    assert popen_kwargs["encoding"] == "UTF-8"
    assert popen_kwargs["stdout"] == subprocess.PIPE
    assert popen_kwargs["stderr"] == subprocess.STDOUT
    assert popen_kwargs["bufsize"] == 1

    dev_command._stream_app_logs.assert_called_once_with(
        first_app,
        popen=log_popen,
        clean_output=False,
    )



from unittest.mock import MagicMock

from briefcase.integrations.virtual_environment import VenvContext


def test_recreate_when_venv_exists(mock_tools, venv_path):
    """Recreate removes existing venv and creates new one."""

    context = VenvContext(mock_tools, venv_path)
    context.exists = MagicMock(return_value=True)
    context.create = MagicMock()

    venv_path.mkdir()
    (venv_path / "pyvenv.cfg").touch()

    assert venv_path.exists()

    context.recreate()
    context.exists.assert_called_once()
    context.create.assert_called_once()

    assert not venv_path.exists()


def test_recreate_when_venv_missing(mock_tools, venv_path):
    """Recreate skips removal and creates new venv."""

    context = VenvContext(mock_tools, venv_path)
    context.exists = MagicMock(return_value=False)
    context.create = MagicMock()

    assert not venv_path.exists()

    context.recreate()
    context.exists.assert_called_once()
    context.create.assert_called_once()

    assert not venv_path.exists()



import subprocess

import pytest

from briefcase.exceptions import BriefcaseCommandError


def test_forward(mock_tools, adb):
    """A port forwarding."""
    # Invoke forward
    adb.forward(5555, 6666)

    # Validate call parameters.
    mock_tools.subprocess.check_output.assert_called_once_with(
        [
            mock_tools.android_sdk.adb_path,
            "-s",
            "exampleDevice",
            "forward",
            "tcp:5555",
            "tcp:6666",
        ],
    )


def test_forward_failure(adb, mock_tools):
    """If port forwarding fails, the error is caught."""
    # Mock out the run command on an adb instance
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        returncode=1, cmd=""
    )
    with pytest.raises(BriefcaseCommandError):
        adb.forward(5555, 6666)


def test_forward_remove(mock_tools, adb):
    """A port forwarding removal."""
    # Invoke forward remove
    adb.forward_remove(5555)

    # Validate call parameters.
    mock_tools.subprocess.check_output.assert_called_once_with(
        [
            mock_tools.android_sdk.adb_path,
            "-s",
            "exampleDevice",
            "forward",
            "--remove",
            "tcp:5555",
        ],
    )


def test_forward_remove_failure(adb, mock_tools):
    """If port forwarding removal fails, the error is caught."""
    # Mock out the run command on an adb instance
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        returncode=1, cmd=""
    )
    with pytest.raises(BriefcaseCommandError):
        adb.forward_remove(5555)


def test_reverse(mock_tools, adb):
    """A port reversing."""
    # Invoke reverse
    adb.reverse(5555, 6666)

    # Validate call parameters.
    mock_tools.subprocess.check_output.assert_called_once_with(
        [
            mock_tools.android_sdk.adb_path,
            "-s",
            "exampleDevice",
            "reverse",
            "tcp:5555",
            "tcp:6666",
        ],
    )


def test_reverse_failure(adb, mock_tools):
    """If port reversing fails, the error is caught."""
    # Mock out the run command on an adb instance
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        returncode=1, cmd=""
    )
    with pytest.raises(BriefcaseCommandError):
        adb.reverse(5555, 6666)


def test_reverse_remove(mock_tools, adb):
    """A port reversing removal."""
    # Invoke reverse remove
    adb.reverse_remove(5555)

    # Validate call parameters.
    mock_tools.subprocess.check_output.assert_called_once_with(
        [
            mock_tools.android_sdk.adb_path,
            "-s",
            "exampleDevice",
            "reverse",
            "--remove",
            "tcp:5555",
        ],
    )


def test_reverse_remove_failure(adb, mock_tools):
    """If port reversing removal fails, the error is caught."""
    # Mock out the run command on an adb instance
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        returncode=1, cmd=""
    )
    with pytest.raises(BriefcaseCommandError):
        adb.reverse_remove(5555)



import pytest

from briefcase.platforms.web.static import StaticWebBuildCommand

from ....utils import create_wheel


@pytest.fixture
def build_command(dummy_console, tmp_path):
    return StaticWebBuildCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


def test_process_wheel(build_command, tmp_path):
    """A wheel can be processed to have CSS content extracted."""

    # Create a wheel with some content
    wheel_filename = create_wheel(
        tmp_path,
        extra_content=[
            # Three CSS files
            (
                "dummy/static/first.css",
                "span {\n  font-color: red;\n  font-size: larger\n}\n",
            ),
            (
                "dummy/static/second.css",
                "div {\n  padding: 10px\n}\n",
            ),
            (
                "dummy/static/deep/third.css",
                "p {\n  color: red\n}\n",
            ),
            # Content in the static file that isn't CSS
            ("dummy/static/explosions.js", "alert('boom!');"),
            # CSS in a location that isn't the static folder
            ("dummy/other.css", "div.other {\n  margin: 10px\n}\n"),
            ("lost.css", "div.lost {\n  margin: 10px\n}\n"),
        ],
    )

    # Collect into inserts dict
    inserts = {}
    build_command._process_wheel(wheelfile=wheel_filename, inserts=inserts)

    # Legacy CSS should be collected into the style.css "css" insert slot
    assert inserts == {
        "static/css/style.css": {
            "css": {
                "dummy 1.2.3 (legacy static CSS: first.css)": (
                    "span {\n  font-color: red;\n  font-size: larger\n}\n"
                ),
                "dummy 1.2.3 (legacy static CSS: second.css)": (
                    "div {\n  padding: 10px\n}\n"
                ),
                "dummy 1.2.3 (legacy static CSS: deep/third.css)": (
                    "p {\n  color: red\n}\n"
                ),
            }
        }
    }


def test_process_wheel_no_content(build_command, tmp_path):
    """A wheel with no resources can be processed."""

    # Create a wheel with some content, but nothing static to be collected
    wheel_filename = create_wheel(
        tmp_path,
        extra_content=[
            # Content in the static file that isn't CSS
            ("dummy/static/explosions.js", "alert('boom!');"),
            # CSS in a location that isn't the static folder.
            ("dummy/other.css", "div.other {\n  margin: 10px\n}\n"),
            ("lost.css", "div.lost {\n  margin: 10px\n}\n"),
        ],
    )

    inserts = {}
    build_command._process_wheel(wheelfile=wheel_filename, inserts=inserts)

    assert inserts == {}


def test_process_wheel_deploy_inserts(build_command, tmp_path):
    """Deploy inserts are collected into the correct insert slot."""
    wheel_filename = create_wheel(
        tmp_path,
        extra_content=[
            ("dummy/deploy/inserts/index.html~header", "<script>alert('hi')</script>"),
            (
                "dummy/deploy/inserts/static/css/style.css~CSS",
                "body { margin: 0; }",
            ),
        ],
    )

    inserts = {}
    build_command._process_wheel(wheelfile=wheel_filename, inserts=inserts)

    # The index.html header insert exists
    assert "index.html" in inserts
    assert "header" in inserts["index.html"]
    contribs = inserts["index.html"]["header"]
    assert any("<script>" in v for v in contribs.values())

    # The CSS insert exists
    assert "static/css/style.css" in inserts
    assert "CSS" in inserts["static/css/style.css"]
    css_contribs = inserts["static/css/style.css"]["CSS"]
    assert any("body { margin: 0; }" in v for v in css_contribs.values())


def test_process_wheel_legacy_css_warn_once(build_command, tmp_path, capsys):
    """Legacy CSS files trigger a single deprecation warning."""
    wheel_filename = create_wheel(
        tmp_path,
        extra_content=[
            ("dummy/static/one.css", "h1 { color: blue; }"),
            ("dummy/static/two.css", "h2 { color: green; }"),
        ],
    )

    inserts = {}
    build_command._process_wheel(wheelfile=wheel_filename, inserts=inserts)

    output = capsys.readouterr().out
    assert (
        "dummy-1.2.3-py3-none-any.whl: legacy '/static' CSS file dummy/static/one.css detected."
        in output
    )
    assert (
        "dummy-1.2.3-py3-none-any.whl: legacy '/static' CSS file dummy/static/two.css detected."
        in output
    )



import sys

import pytest
from packaging.version import Version

from briefcase.exceptions import UnsupportedHostError
from briefcase.platforms.windows.app import WindowsAppCreateCommand


@pytest.fixture
def create_command(dummy_console, tmp_path):
    return WindowsAppCreateCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


@pytest.mark.parametrize("host_os", ["Darwin", "Linux", "WeirdOS"])
def test_unsupported_host_os(create_command, host_os):
    """Error raised for an unsupported OS."""
    create_command.tools.host_os = host_os

    with pytest.raises(
        UnsupportedHostError,
        match=r"Windows applications can only be built on Windows\.",
    ):
        create_command()


@pytest.mark.parametrize("host_arch", ["i686", "ARM64", "wonky"])
def test_unsupported_arch(create_command, host_arch, first_app_config):
    """Internal apps can only be developed on x86-64."""
    create_command.tools.host_os = "Windows"
    create_command.tools.host_arch = host_arch
    create_command.apps["first"] = first_app_config

    with pytest.raises(
        UnsupportedHostError,
        match=f"Windows applications cannot be built on an {host_arch} machine.",
    ):
        create_command.verify_host()


@pytest.mark.parametrize("host_arch", ["i686", "ARM64", "wonky"])
def test_unsupported_arch_external(create_command, host_arch, first_app_config, capsys):
    """External apps can be built on a different architecture, with a warning."""
    create_command.tools.host_os = "Windows"
    create_command.tools.host_arch = host_arch
    create_command.apps["first"] = first_app_config
    first_app_config.external_package_path = "external"

    create_command.verify_host()
    stdout, stderr = capsys.readouterr()
    assert "WARNING: Possible architecture mismatch" in stdout
    assert stderr == ""

    # The warning is not repeated on cloned commands.
    create_command.is_clone = True
    create_command.verify_host()
    assert capsys.readouterr() == ("", "")


def test_supported_arch(create_command):
    """Windows command are allowed to run on x86-64."""
    create_command.tools.host_os = "Windows"
    create_command.tools.host_arch = "AMD64"

    create_command.verify_host()


def test_unsupported_32bit_python(create_command):
    """Windows commands cannot run with 32bit Python."""
    create_command.tools.host_os = "Windows"
    create_command.tools.host_arch = "AMD64"
    create_command.tools.is_32bit_python = True

    with pytest.raises(
        UnsupportedHostError,
        match="Windows applications cannot be built using a 32bit version of Python",
    ):
        create_command()


def test_context(create_command, first_app_config):
    context = create_command.output_format_template_context(first_app_config)
    assert sorted(context.keys()) == [
        "binary_path",
        "guid",
        "install_scope",
        "package_path",
        "version_triple",
    ]


@pytest.mark.parametrize(
    ("version", "version_triple"),
    [
        ("1", "1.0.0"),
        ("1.2", "1.2.0"),
        ("1.2.3", "1.2.3"),
        ("1.2.3.4", "1.2.3"),
        ("1.2.3a4", "1.2.3"),
        ("1.2.3b5", "1.2.3"),
        ("1.2.3rc6", "1.2.3"),
        ("1.2.3.dev7", "1.2.3"),
        ("1.2.3.post8", "1.2.3"),
    ],
)
def test_version_triple(
    create_command, first_app_config, tmp_path, version, version_triple
):
    first_app_config.version = Version(version)
    context = create_command.output_format_template_context(first_app_config)

    assert context["version_triple"] == version_triple


def test_explicit_version_triple(create_command, first_app_config, tmp_path):
    first_app_config.version = "1.2.3a1"
    first_app_config.version_triple = "2.3.4"

    context = create_command.output_format_template_context(first_app_config)

    # Explicit version triple is used.
    assert context["version_triple"] == "2.3.4"


def test_guid(create_command, first_app_config, tmp_path):
    """A predictable GUID will be generated from the bundle."""
    context = create_command.output_format_template_context(first_app_config)

    assert context["guid"] == "d666a4f1-c7b7-52cc-888a-3a35a7cc97e5"


def test_explicit_guid(create_command, first_app_config, tmp_path):
    """If a GUID is explicitly provided, it is used."""
    first_app_config.guid = "e822176f-b755-589f-849c-6c6600f7efb1"
    context = create_command.output_format_template_context(first_app_config)

    # Explicitly provided GUID is used.
    assert context["guid"] == "e822176f-b755-589f-849c-6c6600f7efb1"


@pytest.mark.parametrize(
    ("revision", "micro"),
    [
        # Numerical revision
        (5, "5"),
        # Text revision
        ("5", "5"),
        # Pre-release revision
        ("0a3", "0"),
        ("0b2", "0"),
        ("0rc1", "0"),
    ],
)
def test_support_package_url(
    create_command, revision, micro, first_app_config, tmp_path
):
    """A valid support package URL is created for a support revision."""
    expected_link = (
        f"https://www.python.org/ftp/python"
        f"/{sys.version_info.major}.{sys.version_info.minor}.{micro}"
        f"/python-{sys.version_info.major}.{sys.version_info.minor}.{revision}-embed-amd64.zip"
    )
    assert create_command.support_package_url(revision) == expected_link


def test_default_install_scope(create_command, first_app_config, tmp_path):
    """By default, the installer gives a choice between per user and per machine."""
    context = create_command.output_format_template_context(first_app_config)
    assert context["install_scope"] == "perUserOrMachine"


def test_per_machine_install_scope(create_command, first_app_config, tmp_path):
    """App can be set to have explicit per-machine scope."""
    first_app_config.system_installer = True
    context = create_command.output_format_template_context(first_app_config)
    assert context["install_scope"] == "perMachine"


def test_per_user_install_scope(create_command, first_app_config, tmp_path):
    """App can be set to have explicit per-user scope."""
    first_app_config.system_installer = False
    context = create_command.output_format_template_context(first_app_config)
    assert context["install_scope"] == "perUser"


def test_package_path(create_command, first_app_config, tmp_path):
    """The default package_path is passed as an absolute path."""
    context = create_command.output_format_template_context(first_app_config)
    assert context["package_path"] == str(
        tmp_path / "base_path/build/first-app/windows/app/src"
    )


def test_binary_path(create_command, first_app_config, tmp_path):
    """The default binary_path is passed as a path relative to package_path."""
    context = create_command.output_format_template_context(first_app_config)
    assert context["binary_path"] == "First App.exe"


def test_external(create_command, external_first_app, tmp_path):
    """The package_path and binary_path can be overridden by the user."""
    context = create_command.output_format_template_context(external_first_app)
    assert context["package_path"] == str(tmp_path / "base_path/external/src")
    assert context["binary_path"] == "internal/app.exe"



from textwrap import dedent

import pytest

from briefcase.platforms.windows import txt_to_rtf


@pytest.mark.parametrize(
    ("input", "expected"),
    [
        pytest.param(
            "",
            "{\\rtf1\\ansi\\deff0 {\\fonttbl {\\f0 Courier;}}\n\\par\\line\n}",
            id="empty-doc",
        ),
        pytest.param(
            """\
            Hello World.

            This is a 2 paragraph document
            with a multi-line paragraph.
            """,
            """\
            {\\rtf1\\ansi\\deff0 {\\fonttbl {\\f0 Courier;}}
            Hello World.\x20
            \\par\\line
            This is a 2 paragraph document\x20
            with a multi-line paragraph.\x20
            \\par\\line
            }""",
            id="2-para-document",
        ),
        pytest.param(
            """\
            This is a document that contains bullet points:

             * first, something short
             * then, something longer that needs to
               run onto multiple lines
             * last, something short again

            Then a closing paragraph.
            """,
            """\
            {\\rtf1\\ansi\\deff0 {\\fonttbl {\\f0 Courier;}}
            This is a document that contains bullet points:\x20
            \\par\\line
            \\bullet first, something short\x20
            \\bullet then, something longer that needs to\x20
            run onto multiple lines\x20
            \\bullet last, something short again\x20
            \\par\\line
            Then a closing paragraph.\x20
            \\par\\line
            }""",
            id="bullet-document",
        ),
    ],
)
def test_txt_to_rtf(input, expected):
    """Utility method can convert TXT to RTF format."""
    assert txt_to_rtf(dedent(input)) == dedent(expected)



import pytest

from briefcase.config import AppConfig
from briefcase.platforms.linux.system import LinuxSystemCreateCommand

from ....utils import create_file


@pytest.fixture
def create_command(dummy_console, tmp_path):
    return LinuxSystemCreateCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


@pytest.fixture
def first_app(first_app_config, tmp_path):
    """A fixture for the first app, rolled out on disk."""
    # Specify a system python app for a dummy vendor
    first_app_config.target_vendor = "somevendor"
    first_app_config.target_codename = "surprising"
    first_app_config.target_vendor_base = "basevendor"

    # Targeted Python version
    first_app_config.python_version_tag = "3"

    # Some project-level files.
    create_file(tmp_path / "base_path/LICENSE", "First App License")
    create_file(tmp_path / "base_path/CHANGELOG", "First App Changelog")

    # Make it look like the template has been generated
    bundle_dir = tmp_path / "base_path/build/first-app/somevendor/surprising"

    create_file(bundle_dir / "first-app.1", "First App manpage")

    lib_dir = bundle_dir / "first-app-0.0.1/usr/lib/first-app"
    (lib_dir / "app").mkdir(parents=True, exist_ok=True)
    (lib_dir / "app_packages/firstlib").mkdir(parents=True, exist_ok=True)
    (lib_dir / "app_packages/secondlib").mkdir(parents=True, exist_ok=True)

    # Create some .so files
    # An SO file with different group and world permissions
    (lib_dir / "app/support.so").touch()
    (lib_dir / "app/support.so").chmod(0o775)

    # An SO file with same group and world permissions
    (lib_dir / "app/support_same_perms.so").touch()
    (lib_dir / "app/support_same_perms.so").chmod(0o744)

    # A SO file with both .so and .so.1.0 forms
    (lib_dir / "app_packages/firstlib/first.so").touch()
    (lib_dir / "app_packages/firstlib/first.so.1.0").touch()

    # An SO file with 664 permissions
    (lib_dir / "app_packages/secondlib/second_a.so").touch()
    (lib_dir / "app_packages/secondlib/second_a.so").chmod(0o664)

    (lib_dir / "app_packages/secondlib/second_b.so").touch()

    return first_app_config


@pytest.fixture
def underscore_app(tmp_path):
    """A fixture for an app with an underscore in the name, rolled out on disk."""
    app_config = AppConfig(
        app_name="underscore_app",
        project_name="Underscore Project",
        formal_name="Underscore App",
        author="Megacorp",
        author_email="maintainer@example.com",
        url="https://example.com/underscore_app",
        bundle="com.example",
        version="0.0.1",
        description="The first simple app \\ demonstration",
        sources=["src/underscore_app"],
        license={"file": "LICENSE"},
    )

    # Specify a system python app for a dummy vendor
    app_config.target_vendor = "somevendor"
    app_config.target_codename = "surprising"
    app_config.target_vendor_base = "basevendor"

    # Targeted Python version
    app_config.python_version_tag = "3"

    # Some project-level files.
    create_file(tmp_path / "base_path/LICENSE", "Underscore App License")
    create_file(tmp_path / "base_path/CHANGELOG", "Underscore App Changelog")

    # Make it look like the template has been generated
    bundle_dir = tmp_path / "base_path/build/underscore_app/somevendor/surprising"

    lib_dir = bundle_dir / "underscore_app-0.0.1/usr/lib/underscore_app"
    (lib_dir / "app").mkdir(parents=True, exist_ok=True)

    return app_config



import shutil
import subprocess
import uuid
from pathlib import Path
from unittest import mock

import pytest

from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.subprocess import json_parser

from .....utils import create_file


def test_resume_notarize_app(
    package_command,
    first_app_with_binaries,
    sekrit_identity,
    tmp_path,
    sleep_zero,
):
    """Notarization of an app bundle can be resumed."""
    # Select a codesigning identity
    package_command.select_identity.return_value = sekrit_identity

    # Mock the creation of the ditto archive
    package_command.ditto_archive = mock.MagicMock()

    # Mock the return values of subprocesses
    submission_id = str(uuid.uuid4())
    package_command.tools.subprocess.parse_output.side_effect = [
        # notarytool history returns list with the app's submission ID
        {
            "history": [
                {
                    "id": str(uuid.uuid4()),
                    "name": "Other App-1.2.3.dmg",
                    "status": "Accepted",
                },
                {
                    "id": submission_id,
                    "name": "First App.app.zip",
                    "status": "In Progress",
                },
            ]
        },
        # notarytool log; 2 failures, then a successful result.
        subprocess.CalledProcessError(
            returncode=69,
            cmd=["xcrun", "notarytool", "log"],
        ),
        subprocess.CalledProcessError(
            returncode=69,
            cmd=["xcrun", "notarytool", "log"],
        ),
        {"status": "Accepted"},
    ]

    # Resume notarizaton. Use the base command's interface to ensure the full cleanup
    # process is tested.
    package_command._package_app(
        first_app_with_binaries,
        update=False,
        packaging_format="zip",
        identity=sekrit_identity.id,
        submission_id=submission_id,
    )

    # Identity selection excluded adhoc identities
    package_command.select_identity.assert_called_once_with(
        identity=sekrit_identity.id,
        allow_adhoc=False,
    )

    # The calls to notarization tools were made
    assert package_command.tools.subprocess.parse_output.mock_calls == [
        # Retrieve notarization history to verify the submission ID
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "history",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                "--output-format",
                "json",
            ],
        ),
        # Check status 3 times
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "log",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                submission_id,
            ],
            quiet=1,
        ),
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "log",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                submission_id,
            ],
            quiet=1,
        ),
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "log",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                submission_id,
            ],
            quiet=1,
        ),
    ]

    # Notarization is complete; we can staple.
    package_command.tools.subprocess.run.assert_called_once_with(
        [
            "xcrun",
            "stapler",
            "staple",
            tmp_path / "base_path/build/first-app/macos/app/First App.app",
        ],
        check=True,
    )

    # As this is a .zip distribution, there's a finalization step to create the actual
    # distribution artefact with ditto.
    package_command.ditto_archive.assert_called_once_with(
        tmp_path / "base_path/build/first-app/macos/app/First App.app",
        tmp_path / "base_path/dist/First App-0.0.1.app.zip",
    )


def test_resume_notarize_dmg(
    package_command,
    first_app_with_binaries,
    sekrit_identity,
    tmp_path,
    sleep_zero,
):
    """Notarization of a DMG can be resumed."""
    # Create a pre-existing distribution artefact
    create_file(
        tmp_path / "base_path/dist/First App-0.0.1.dmg",
        "distribution file",
    )

    # Select a codesigning identity
    package_command.select_identity.return_value = sekrit_identity

    # Mock the creation of the ditto archive
    package_command.ditto_archive = mock.MagicMock()

    # Mock the return values of subprocesses
    submission_id = str(uuid.uuid4())
    package_command.tools.subprocess.parse_output.side_effect = [
        # notarytool history returns list with the app's submission ID
        {
            "history": [
                {
                    "id": str(uuid.uuid4()),
                    "name": "Other App-1.2.3.dmg",
                    "status": "Accepted",
                },
                {
                    "id": submission_id,
                    "name": "First App-0.0.1.dmg",
                    "status": "In Progress",
                },
            ]
        },
        # notarytool log; 2 failures, then a successful result.
        subprocess.CalledProcessError(
            returncode=69,
            cmd=["xcrun", "notarytool", "log"],
        ),
        subprocess.CalledProcessError(
            returncode=69,
            cmd=["xcrun", "notarytool", "log"],
        ),
        {"status": "Accepted"},
    ]

    # Resume notarizaton
    package_command._package_app(
        first_app_with_binaries,
        update=False,
        packaging_format="dmg",
        identity=sekrit_identity.id,
        submission_id=submission_id,
    )

    # Identity selection excluded adhoc identities
    package_command.select_identity.assert_called_once_with(
        identity=sekrit_identity.id,
        allow_adhoc=False,
    )

    # The calls to notarization tools were made
    assert package_command.tools.subprocess.parse_output.mock_calls == [
        # Retrieve notarization history to verify the submission ID
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "history",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                "--output-format",
                "json",
            ],
        ),
        # Check status 3 times
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "log",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                submission_id,
            ],
            quiet=1,
        ),
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "log",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                submission_id,
            ],
            quiet=1,
        ),
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "log",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                submission_id,
            ],
            quiet=1,
        ),
    ]

    # Notarization is complete; we can staple.
    package_command.tools.subprocess.run.assert_called_once_with(
        [
            "xcrun",
            "stapler",
            "staple",
            tmp_path / "base_path/dist/First App-0.0.1.dmg",
        ],
        check=True,
    )

    # The distribution artefact already exists, so there's no call to ditto.
    package_command.ditto_archive.assert_not_called()


def test_resume_notarize_pkg(
    package_command,
    first_app_with_binaries,
    sekrit_identity,
    sekrit_installer_identity,
    tmp_path,
    sleep_zero,
):
    """Notarization of a PKG installer can be resumed."""
    # Create a pre-existing distribution artefact
    create_file(
        tmp_path / "base_path/dist/First App-0.0.1.pkg",
        "distribution file",
    )

    # 2 calls are made to determine identity - the app identity, then the installer identity.
    package_command.select_identity.side_effect = [
        sekrit_identity,
        sekrit_installer_identity,
    ]

    # Mock the creation of the ditto archive
    package_command.ditto_archive = mock.MagicMock()

    # Mock the return values of subprocesses
    submission_id = str(uuid.uuid4())
    package_command.tools.subprocess.parse_output.side_effect = [
        # notarytool history returns list with the app's submission ID
        {
            "history": [
                {
                    "id": str(uuid.uuid4()),
                    "name": "Other App-1.2.3.dmg",
                    "status": "Accepted",
                },
                {
                    "id": submission_id,
                    "name": "First App-0.0.1.pkg",
                    "status": "In Progress",
                },
            ]
        },
        # notarytool log; 2 failures, then a successful result.
        subprocess.CalledProcessError(
            returncode=69,
            cmd=["xcrun", "notarytool", "log"],
        ),
        subprocess.CalledProcessError(
            returncode=69,
            cmd=["xcrun", "notarytool", "log"],
        ),
        {"status": "Accepted"},
    ]

    # Resume notarizaton
    package_command._package_app(
        first_app_with_binaries,
        update=False,
        packaging_format="pkg",
        identity=sekrit_identity.id,
        installer_identity=sekrit_installer_identity.id,
        submission_id=submission_id,
    )

    # Identity selection excluded adhoc identities, but also confirmed notarization identity
    assert package_command.select_identity.mock_calls == [
        mock.call(
            identity=sekrit_identity.id,
            allow_adhoc=False,
        ),
        mock.call(
            identity=sekrit_installer_identity.id,
            app_identity=sekrit_identity,
        ),
    ]

    # The calls to notarization tools were made
    assert package_command.tools.subprocess.parse_output.mock_calls == [
        # Retrieve notarization history to verify the submission ID
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "history",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                "--output-format",
                "json",
            ],
        ),
        # Check status 3 times
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "log",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                submission_id,
            ],
            quiet=1,
        ),
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "log",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                submission_id,
            ],
            quiet=1,
        ),
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "log",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                submission_id,
            ],
            quiet=1,
        ),
    ]

    # Notarization is complete; we can staple.
    package_command.tools.subprocess.run.assert_called_once_with(
        [
            "xcrun",
            "stapler",
            "staple",
            tmp_path / "base_path/dist/First App-0.0.1.pkg",
        ],
        check=True,
    )

    # The distribution artefact already exists, so there's no call to ditto.
    package_command.ditto_archive.assert_not_called()


def test_resume_notarize_app_artefact_missing(
    package_command,
    first_app_with_binaries,
    sekrit_identity,
    tmp_path,
):
    """If the app binary doesn't exist, notarization of an app cannot be resumed."""
    # Delete the actual binary
    shutil.rmtree(tmp_path / "base_path/build/first-app/macos/app/First App.app")

    # Mock the command factory to prevent the binary from being rebuilt as part of
    # package dependencies.
    package_command._command_factory = mock.MagicMock()

    # Attempting to resume notarization when there's no pre-existing artefact raises an
    # error.
    with pytest.raises(
        BriefcaseCommandError,
        match=(
            r"Notarization cannot be resumed, as the notarization artefact "
            r"associated with this app \(.*First App.app\) does not exist."
        ),
    ):
        package_command._package_app(
            first_app_with_binaries,
            update=False,
            packaging_format="zip",
            identity=sekrit_identity.id,
            submission_id=str(uuid.uuid4()),
        )


def test_resume_notarize_app_dist_artefact_exists(
    package_command,
    first_app_with_binaries,
    sekrit_identity,
    tmp_path,
    sleep_zero,
):
    """A *distribution* artefact of an app bundle will be cleaned up as part of resuming
    notarization."""
    # Create a pre-existing distribution artefact
    create_file(
        tmp_path / "base_path/dist/First App-0.0.1.app.zip",
        "distribution file",
    )

    # Select a codesigning identity
    package_command.select_identity.return_value = sekrit_identity

    # Mock the creation of the ditto archive
    package_command.ditto_archive = mock.MagicMock()

    # Mock the return values of subprocesses
    submission_id = str(uuid.uuid4())
    package_command.tools.subprocess.parse_output.side_effect = [
        # notarytool history returns list with the app's submission ID
        {
            "history": [
                {
                    "id": str(uuid.uuid4()),
                    "name": "Other App-1.2.3.dmg",
                    "status": "Accepted",
                },
                {
                    "id": submission_id,
                    "name": "First App.app.zip",
                    "status": "In Progress",
                },
            ]
        },
        # notarytool log; 2 failures, then a successful result.
        subprocess.CalledProcessError(
            returncode=69,
            cmd=["xcrun", "notarytool", "log"],
        ),
        subprocess.CalledProcessError(
            returncode=69,
            cmd=["xcrun", "notarytool", "log"],
        ),
        {"status": "Accepted"},
    ]

    # Resume notarizaton. Use the base command's interface to ensure the full cleanup
    # process is tested.
    package_command._package_app(
        first_app_with_binaries,
        update=False,
        packaging_format="zip",
        identity=sekrit_identity.id,
        submission_id=submission_id,
    )

    # Identity selection excluded adhoc identities
    package_command.select_identity.assert_called_once_with(
        identity=sekrit_identity.id,
        allow_adhoc=False,
    )

    # The calls to notarization tools were made
    assert package_command.tools.subprocess.parse_output.mock_calls == [
        # Retrieve notarization history to verify the submission ID
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "history",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                "--output-format",
                "json",
            ],
        ),
        # Check status 3 times
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "log",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                submission_id,
            ],
            quiet=1,
        ),
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "log",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                submission_id,
            ],
            quiet=1,
        ),
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "log",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                submission_id,
            ],
            quiet=1,
        ),
    ]

    # Notarization is complete; we can staple.
    package_command.tools.subprocess.run.assert_called_once_with(
        [
            "xcrun",
            "stapler",
            "staple",
            tmp_path / "base_path/build/first-app/macos/app/First App.app",
        ],
        check=True,
    )

    # As this is a .zip distribution, there's a finalization step to create the actual
    # distribution artefact with ditto. This file won't actually be created though, as
    # we're mocking ditto. However, as a result of resuming with a pre-existing
    # distribution artefact, the pre-existing artefact will have been deleted.
    package_command.ditto_archive.assert_called_once_with(
        tmp_path / "base_path/build/first-app/macos/app/First App.app",
        tmp_path / "base_path/dist/First App-0.0.1.app.zip",
    )

    assert not (tmp_path / "base_path/dist/First App-0.0.1.app.zip").exists()


def test_resume_notarize_artefact_missing(
    package_command,
    first_app_with_binaries,
    sekrit_identity,
):
    """If the distribution artefact doesn't exist, notarization cannot be resumed."""

    # Attempting to resume notarization when there's no pre-existing artefact raises an
    # error.
    with pytest.raises(
        BriefcaseCommandError,
        match=(
            r"Notarization cannot be resumed, as the notarization artefact "
            r"associated with this app \(.*First App-0.0.1.dmg\) does not exist."
        ),
    ):
        package_command._package_app(
            first_app_with_binaries,
            update=False,
            packaging_format="dmg",
            identity=sekrit_identity.id,
            submission_id=str(uuid.uuid4()),
        )


def test_invalid_submission_id(
    package_command,
    first_app_with_binaries,
    sekrit_identity,
    tmp_path,
):
    """If an invalid submission ID is provided, an error is raised."""
    # Create a pre-existing distribution artefact
    create_file(
        tmp_path / "base_path/dist/First App-0.0.1.dmg",
        "distribution file",
    )

    # Select a codesigning identity
    package_command.select_identity.return_value = sekrit_identity

    # Mock the return values of subprocesses
    submission_id = str(uuid.uuid4())
    package_command.tools.subprocess.parse_output.side_effect = [
        # notarytool history returns list with submission IDs that don't
        # match the expected submission ID
        {
            "history": [
                {
                    "id": str(uuid.uuid4()),
                    "name": "Other App-1.2.3.dmg",
                    "status": "Accepted",
                },
                {
                    "id": str(uuid.uuid4()),
                    "name": "First App-0.0.1.dmg",
                    "status": "In Progress",
                },
            ]
        },
    ]

    # Attempt to resume notarizaton. The submission ID will be rejected.
    with pytest.raises(
        BriefcaseCommandError,
        match=rf"{submission_id} is not a known submission ID for this identity.",
    ):
        package_command._package_app(
            first_app_with_binaries,
            update=False,
            packaging_format="dmg",
            identity=sekrit_identity.id,
            submission_id=submission_id,
        )

    # The history has been retrieved.
    package_command.tools.subprocess.parse_output.assert_called_once_with(
        json_parser,
        [
            "xcrun",
            "notarytool",
            "history",
            "--keychain-profile",
            "briefcase-macOS-DEADBEEF",
            "--output-format",
            "json",
        ],
    )

    # No staple attempt is made
    package_command.tools.subprocess.run.assert_not_called()


def test_invalid_notarization_identity(
    package_command,
    first_app_with_binaries,
    sekrit_identity,
    tmp_path,
):
    """If an invalid notarization identity is provided, an error is raised."""
    # Create a pre-existing distribution artefact
    create_file(
        tmp_path / "base_path/dist/First App-0.0.1.dmg",
        "distribution file",
    )

    # Select a codesigning identity
    package_command.select_identity.return_value = sekrit_identity

    # Mock the return values of subprocesses
    submission_id = str(uuid.uuid4())
    package_command.tools.subprocess.parse_output.side_effect = [
        # notarytool history returns an error
        subprocess.CalledProcessError(69, cmd="xcrun notarytool history")
    ]

    # Attempt to resume notarizaton. The submission ID will not be validated
    with pytest.raises(
        BriefcaseCommandError,
        match=(
            r"Unable to invoke notarytool to determine validity of submission ID.\n"
            r"Are you sure this is the identity that was used to notarize the app\?"
        ),
    ):
        package_command._package_app(
            first_app_with_binaries,
            update=False,
            packaging_format="dmg",
            identity=sekrit_identity.id,
            submission_id=submission_id,
        )

    # The history has been retrieved.
    package_command.tools.subprocess.parse_output.assert_called_once_with(
        json_parser,
        [
            "xcrun",
            "notarytool",
            "history",
            "--keychain-profile",
            "briefcase-macOS-DEADBEEF",
            "--output-format",
            "json",
        ],
    )

    # No staple attempt is made
    package_command.tools.subprocess.run.assert_not_called()


@pytest.mark.parametrize(
    "dist_filename",
    [
        "First App-0.0.1.app.zip",
        "First App-0.0.1.dmg",
        "First App-0.0.1.pkg",
    ],
)
def test_filename_mismatch(
    package_command,
    first_app_with_binaries,
    sekrit_identity,
    tmp_path,
    dist_filename,
):
    """If the filename on a notarization submission doesn't match the current artefact,
    an error is raised."""
    # Create a pre-existing distribution artefact
    create_file(
        tmp_path / "base_path/dist" / dist_filename,
        "distribution file",
    )

    # Select a codesigning identity
    package_command.select_identity.return_value = sekrit_identity

    # Mock the return values of subprocesses
    submission_id = str(uuid.uuid4())
    package_command.tools.subprocess.parse_output.side_effect = [
        # notarytool history returns list with a submission ID that matches,
        # but the filename won't.
        {
            "history": [
                {
                    "id": str(uuid.uuid4()),
                    "name": "Other App-1.2.3.dmg",
                    "status": "Accepted",
                },
                {
                    "id": submission_id,
                    "name": "Third App-0.0.1.dmg",
                    "status": "In Progress",
                },
            ]
        },
    ]

    # Attempt to resume notarizaton. The submission ID will be rejected.
    with pytest.raises(
        BriefcaseCommandError,
        match=(
            rf"{submission_id} is not a submission ID for this project. "
            r"It notarizes a file named Third App-0.0.1.dmg"
        ),
    ):
        package_command._package_app(
            first_app_with_binaries,
            update=False,
            packaging_format=Path(dist_filename).suffix[1:],
            identity=sekrit_identity.id,
            submission_id=submission_id,
        )

    # The history has been retrieved.
    package_command.tools.subprocess.parse_output.assert_called_once_with(
        json_parser,
        [
            "xcrun",
            "notarytool",
            "history",
            "--keychain-profile",
            "briefcase-macOS-DEADBEEF",
            "--output-format",
            "json",
        ],
    )

    # No staple attempt is made
    package_command.tools.subprocess.run.assert_not_called()


@pytest.mark.parametrize(
    ("response", "error"),
    [
        # A response with an unknown status code.
        (
            {"status": "wibble"},
            r"Unexpected notarization status: wibble",
        ),
        # A bare bones response. This isn't something we've seen in practice, but we
        # protect against it anyway.
        (
            {"status": "Invalid"},
            r"Notarization was rejected: No details provided",
        ),
        # A failure with no issues provided.
        (
            {"status": "Invalid", "statusSummary": "Bad mojo"},
            r"Notarization was rejected: Bad mojo",
        ),
        # A failure with an empty list of issues.
        (
            {"status": "Invalid", "statusSummary": "Bad mojo", "issues": []},
            r"Notarization was rejected: Bad mojo",
        ),
        # A failure with a list of minimalist issues
        (
            {
                "status": "Invalid",
                "statusSummary": "Bad mojo",
                "issues": [
                    {
                        "path": "foo/bar",
                        "message": "This isn't good",
                    },
                    {
                        "path": "pork/ham",
                        "message": "This is fairly bad",
                    },
                ],
            },
            (
                r"Notarization was rejected: Bad mojo\n"
                r"\n"
                r"    \* \(\?\) foo/bar \[unknown architecture\]\n"
                r"      This isn't good\n"
                r"      \(No additional help available\)\n"
                r"\n"
                r"    \* \(\?\) pork/ham \[unknown architecture\]\n"
                r"      This is fairly bad\n"
                r"      \(No additional help available\)"
            ),
        ),
        # A failure with a fully populated list of issues
        (
            {
                "status": "Invalid",
                "statusSummary": "Bad mojo",
                "issues": [
                    {
                        "severity": "Error",
                        "path": "foo/bar",
                        "architecture": "arm64",
                        "message": "This isn't good",
                        "docUrl": "https://example.com/error",
                    },
                    {
                        "severity": "Warning",
                        "path": "pork/ham",
                        "architecture": "arm64",
                        "message": "This is fairly bad",
                        "docUrl": "https://example.com/warning",
                    },
                ],
            },
            (
                r"Notarization was rejected: Bad mojo\n"
                r"\n"
                r"    \* \(Error\) foo/bar \[arm64\]\n"
                r"      This isn't good\n"
                r"      https://example.com/error\n"
                r"\n"
                r"    \* \(Warning\) pork/ham \[arm64\]\n"
                r"      This is fairly bad\n"
                r"      https://example.com/warning"
            ),
        ),
    ],
)
def test_notarization_rejected(
    package_command,
    first_app_with_binaries,
    sekrit_identity,
    tmp_path,
    sleep_zero,
    response,
    error,
):
    """If notarization is rejected, an error is raised."""
    # Create a pre-existing distribution artefact
    create_file(
        tmp_path / "base_path/dist/First App-0.0.1.dmg",
        "distribution file",
    )

    # Select a codesigning identity
    package_command.select_identity.return_value = sekrit_identity

    # Mock the return values of subprocesses
    submission_id = str(uuid.uuid4())
    package_command.tools.subprocess.parse_output.side_effect = [
        # notarytool history returns list with the app's submission ID
        {
            "history": [
                {
                    "id": str(uuid.uuid4()),
                    "name": "Other App-1.2.3.dmg",
                    "status": "Accepted",
                },
                {
                    "id": submission_id,
                    "name": "First App-0.0.1.dmg",
                    "status": "In Progress",
                },
            ]
        },
        # notarytool log; 2 failures, then a successful result.
        subprocess.CalledProcessError(
            returncode=69,
            cmd=["xcrun", "notarytool", "log"],
        ),
        subprocess.CalledProcessError(
            returncode=69,
            cmd=["xcrun", "notarytool", "log"],
        ),
        response,
    ]

    # Resume notarizaton
    with pytest.raises(BriefcaseCommandError, match=error):
        package_command._package_app(
            first_app_with_binaries,
            update=False,
            packaging_format="dmg",
            identity=sekrit_identity.id,
            submission_id=submission_id,
        )

    # The calls to notarization tools were made
    assert package_command.tools.subprocess.parse_output.mock_calls == [
        # Retrieve notarization history to verify the submission ID
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "history",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                "--output-format",
                "json",
            ],
        ),
        # Check status 3 times
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "log",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                submission_id,
            ],
            quiet=1,
        ),
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "log",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                submission_id,
            ],
            quiet=1,
        ),
        mock.call(
            json_parser,
            [
                "xcrun",
                "notarytool",
                "log",
                "--keychain-profile",
                "briefcase-macOS-DEADBEEF",
                submission_id,
            ],
            quiet=1,
        ),
    ]

    # No staple attempt is made.
    package_command.tools.subprocess.run.assert_not_called()



import pytest

from briefcase.config import AppConfig
from briefcase.exceptions import BriefcaseCommandError


@pytest.fixture
def first_app():
    return AppConfig(
        app_name="first",
        bundle="com.example",
        version="0.0.1",
        description="The first simple app",
        sources=["src/first"],
        license={"file": "LICENSE"},
    )


@pytest.fixture
def second_app():
    return AppConfig(
        app_name="second",
        bundle="com.example",
        version="0.0.2",
        description="The second simple app",
        sources=["src/second"],
        license={"file": "LICENSE"},
    )


def test_resolve_all_apps(base_command, first_app, second_app):
    """If no app or app_name is provided, all apps are returned."""
    base_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    result = base_command.resolve_apps()

    assert result == {"first": first_app, "second": second_app}


def test_resolve_by_app_name(base_command, first_app, second_app):
    """If app_name is provided, only that app is returned."""
    base_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    result = base_command.resolve_apps(app_name="first")

    assert result == {"first": first_app}


def test_resolve_by_app_config(base_command, first_app, second_app):
    """If an app config is provided, only that app is returned."""
    base_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    result = base_command.resolve_apps(app=first_app)

    assert result == {"first": first_app}


def test_resolve_invalid_app_name(base_command, first_app):
    """If an invalid app_name is provided, an error is raised."""
    base_command.apps = {
        "first": first_app,
    }

    with pytest.raises(
        BriefcaseCommandError,
        match=r"App 'invalid' does not exist in this project.",
    ):
        base_command.resolve_apps(app_name="invalid")


def test_app_name_takes_precedence_over_app(base_command, first_app, second_app):
    """If both app_name and app are provided, app_name takes precedence."""
    base_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    result = base_command.resolve_apps(app_name="first", app=second_app)

    assert result == {"first": first_app}



import pytest

from briefcase.channels import (
    get_publication_channel,
    get_publication_channels,
)
from briefcase.channels.appstore import AppStorePublicationChannel
from briefcase.channels.base import (
    BasePublicationChannel,
    PublishCommandAPI,
)
from briefcase.channels.playstore import PlayStorePublicationChannel
from briefcase.exceptions import BriefcaseCommandError


def test_publish_command_api_is_runtime_checkable():
    """An object with the required attributes satisfies PublishCommandAPI."""

    class Conforming:
        console = None
        tools = None
        dist_path = None

        def distribution_path(self, app):
            pass

    assert isinstance(Conforming(), PublishCommandAPI)


def test_publish_command_api_rejects_incomplete():
    """An object missing required attributes does not satisfy PublishCommandAPI."""

    class Missing:
        console = None

    assert not isinstance(Missing(), PublishCommandAPI)


def test_get_publication_channels_no_match():
    """An unregistered platform/format combination returns an empty dict."""
    assert get_publication_channels("tester", "dummy") == {}


def test_appstore_channel():
    """App Store placeholder channel has the expected interface."""
    channel = AppStorePublicationChannel()
    assert isinstance(channel, BasePublicationChannel)
    assert channel.name == "appstore"


def test_playstore_channel():
    """Play Store placeholder channel has the expected interface."""
    channel = PlayStorePublicationChannel()
    assert isinstance(channel, BasePublicationChannel)
    assert channel.name == "playstore"


@pytest.mark.parametrize(
    ("channel_class", "match"),
    [
        (AppStorePublicationChannel, "iOS App Store is not yet implemented"),
        (PlayStorePublicationChannel, "Google Play Store is not yet implemented"),
    ],
)
def test_placeholder_channel_raises(channel_class, match):
    """Placeholder channels raise an error when publish_app is called."""
    channel = channel_class()
    with pytest.raises(BriefcaseCommandError, match=match):
        channel.publish_app(app=None, command=None)


@pytest.mark.parametrize(
    ("platform", "output_format", "channel", "channel_class"),
    [
        ("iOS", "Xcode", "appstore", AppStorePublicationChannel),
        ("android", "gradle", "playstore", PlayStorePublicationChannel),
    ],
)
def test_get_publication_channels_discovery(
    platform,
    output_format,
    channel,
    channel_class,
):
    """Built-in channels are discovered via entry points."""
    channels = get_publication_channels(platform, output_format)
    assert channels[channel] is channel_class


def test_get_publication_channel():
    """A channel can be retrieved by name."""
    assert isinstance(
        get_publication_channel("appstore", "iOS", "Xcode"),
        AppStorePublicationChannel,
    )


@pytest.mark.parametrize(
    ("platform", "output_format", "channel"),
    [
        # Completely unknown
        ("flat", "wrapped", "something"),
        # Known platform, but unknown format
        ("iOS", "Xcode", "something"),
        # Known format on a different platform
        ("iOS", "Xcode", "playstore"),
    ],
)
def test_get_publication_channel_unknown(platform, output_format, channel):
    """Requesting an unknown channel raises an error."""
    with pytest.raises(
        BriefcaseCommandError,
        match=f"Unknown publication channel: {channel}",
    ):
        get_publication_channel(channel, platform, output_format)



import re
import subprocess
import sys
from unittest.mock import MagicMock

import pytest

from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.virtual_environment import VenvContext


def test_create(mock_tools, venv_path):
    """Create creates a venv at the specified path."""
    context = VenvContext(mock_tools, venv_path)

    # Mock the methods that create() calls
    context.update_core_tools = MagicMock()
    mock_tools.subprocess.run = MagicMock()

    # Call create
    context.create()

    # Verify subprocess.run was called with correct arguments
    mock_tools.subprocess.run.assert_called_once_with(
        [sys.executable, "-m", "venv", venv_path],
        check=True,
    )

    # Verify update_core_tools was called
    context.update_core_tools.assert_called_once()


def test_create_subprocess_fails(mock_tools, venv_path):
    """Create raises BriefcaseCommandError if subprocess fails."""
    context = VenvContext(mock_tools, venv_path)

    context.update_core_tools = MagicMock()

    mock_tools.subprocess.run = MagicMock(
        side_effect=subprocess.CalledProcessError(returncode=1, cmd="venv")
    )

    escaped_venv_path = re.escape(
        str(venv_path)
    )  # Escape backslashes for regex match, windows backslashes

    with pytest.raises(
        BriefcaseCommandError,
        match=f"Failed to create virtual environment at {escaped_venv_path}",
    ):
        context.create()

    mock_tools.subprocess.run.assert_called_once_with(
        [sys.executable, "-m", "venv", venv_path],
        check=True,
    )

    context.update_core_tools.assert_not_called()



from __future__ import annotations

import concurrent
import email
import hashlib
import os
import pathlib
import plistlib
import subprocess
from pathlib import Path

from briefcase.exceptions import BriefcaseCommandError

CORETYPES_PATH = "/System/Library/CoreServices/CoreTypes.bundle/Contents/Info.plist"


def is_mach_o_binary(path: Path):  # pragma: no-cover-if-is-windows
    """Determine if the file at the given path is a Mach-O binary.

    :param path: The path to check
    :returns: True if the file at the given location is a Mach-O binary.
    """
    # A binary is any file that is executable, or has a suffix from a known list
    if os.access(path, os.X_OK) or path.suffix.lower() in {".dylib", ".o", ".so", ""}:
        # File is a binary; read the file magic to determine if it's Mach-O.
        with path.open("rb") as f:
            magic = f.read(4)
            return magic in (
                b"\xca\xfe\xba\xbe",
                b"\xcf\xfa\xed\xfe",
                b"\xce\xfa\xed\xfe",
                b"\xbe\xba\xfe\xca",
                b"\xfe\xed\xfa\xcf",
                b"\xfe\xed\xfa\xce",
            )
    else:
        # Not a binary
        return False


def sha256_file_digest(path: Path) -> str:
    """Compute a sha256 checksum digest of a file.

    :param path: The file to digest.
    :returns: A sha256 hex digest for the file.
    """
    with path.open("rb") as f:
        file_hash = hashlib.sha256()
        while chunk := f.read(8192):
            file_hash.update(chunk)

    return file_hash.hexdigest()


class AppPackagesMergeMixin:
    # A mixin containing the utilities to merge independent platform-specific
    # app_packages folders into a single "fat" app_packages folder.
    # This is currently only used by macOS, but it *could* be required on iOS
    # if they ever re-introduce multiple on-device architectures.

    def find_binary_packages(
        self,
        install_path: Path,
        universal_suffix: str | None = None,
        other_suffix: str | None = None,
    ) -> list[tuple[str, str]]:
        """Find the packages that have been installed that have binary components.

        :param install_path: The path into which packages have been installed.
        :param universal_suffix: The tag suffix that indicates a universal wheel.
        :param other_suffix: The tag suffix for the *other* architecture. Presence of
            this architecture implies a universal wheel, even if it isn't tagged as
            such.
        :returns: A list of (package name, version) tuples, describing the packages that
            are in the install path that are non-universal and non-pure.
        """
        binary_packages = []
        for distinfo in install_path.glob("*.dist-info"):
            # Read the WHEEL file in the dist-info folder.
            # Use this to determine if the wheel is "pure", and the tag
            # for the wheel.
            with (distinfo / "WHEEL").open("r", encoding="utf-8") as f:
                wheel_data = email.message_from_string(f.read())
                is_purelib = wheel_data.get("Root-Is-Purelib", "false") == "true"
                tags = wheel_data.get_all("Tag")

            # If the wheel is pure, it's not a binary package
            if is_purelib:
                continue

            # If the tag ends with the universal tag, the binary package can be used on
            # all targets and doesn't need additional processing.
            if universal_suffix and any(tag.endswith(universal_suffix) for tag in tags):
                continue

            # If the wheel is tagged with the *other* architecture, it's a universal
            # wheel even though it isn't tagged as such
            if other_suffix and any(tag.endswith(other_suffix) for tag in tags):
                continue

            # The wheel is a single platform binary wheel.
            # Read the package metadata to determine what needs to be installed.
            with (distinfo / "METADATA").open("r", encoding="utf-8") as f:
                metadata = email.message_from_string(f.read())
                binary_packages.append((metadata["Name"], metadata["Version"]))

        return binary_packages

    def ensure_thin_binary(self, path: Path, arch: str):
        """Ensure that a binary is thin, targeting a given architecture.

        If the library is already thin, it is left as-is.

        :param path: The library file to process.
        :param arch: The architecture that should be preserved.
        """
        try:
            output = self.tools.subprocess.check_output(
                ["lipo", "-info", path],
            )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                f"Unable to inspect architectures in {path}"
            ) from e
        else:
            if output.startswith("Non-fat file: "):
                self.console.verbose(f"{path} is already thin.")
            elif output.startswith("Architectures in the fat file: "):
                architectures = set(output.strip().split(":")[-1].strip().split(" "))
                if arch in architectures:
                    self.console.verbose(f"Thinning {path}")
                    try:
                        thin_lib_path = path.parent / f"{path.name}.{arch}"
                        self.tools.subprocess.run(
                            [
                                "lipo",
                                "-thin",
                                arch,
                                "-output",
                                thin_lib_path,
                                path,
                            ],
                            check=True,
                        )
                    except subprocess.CalledProcessError as e:
                        raise BriefcaseCommandError(
                            f"Unable to create thin binary from {path}"
                        ) from e
                    else:
                        # Having extracted the single architecture into a temporary
                        # file, replace the original with the thin version.
                        self.tools.shutil.move(thin_lib_path, path)
                else:
                    raise BriefcaseCommandError(
                        f"{path} does not contain a {arch} slice."
                    )
            else:
                raise BriefcaseCommandError(
                    f"Unable to determine architectures in {path}"
                )

    def lipo_dylib(self, relative_path: Path, target_path: Path, sources: list[Path]):
        """Create a fat library by invoking lipo on multiple source libraries.

        :param relative_path: The path fragment for the dylib, relative to the root
        :param target_path: The root location where the fat library will be written
        :param sources: A list of root locations providing single platform libraries.
        """
        self.console.verbose(f"Creating fat library {relative_path}")

        try:
            # Ensure the directory where the library will be written exists.
            (target_path / relative_path).parent.mkdir(exist_ok=True, parents=True)

            # Add all the constructed source paths. If the original binary is universal,
            # or the binary is only needed on *some* platforms (e.g., libjpeg isn't
            # included in the x86_64 Pillow wheel), the source won't exist, so only
            # merge sources that actually exist. lipo allows creating a "fat"
            # single-platform binary; it's effectively a copy.
            self.tools.subprocess.run(
                [
                    "lipo",
                    "-create",
                    "-output",
                    target_path / relative_path,
                ]
                + [
                    source_path / relative_path
                    for source_path in sources
                    if (source_path / relative_path).is_file()
                ],
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                f"Unable to create fat library for {relative_path}"
            ) from e

    def thin_app_packages(
        self,
        app_packages: Path,
        arch: str,
    ):
        """Ensure that all the dylibs in a given app_packages folder are thin."""
        dylibs = []
        for source_path in app_packages.glob("**/*"):
            if not source_path.is_dir() and is_mach_o_binary(source_path):
                dylibs.append(source_path)

        # Call lipo on each dylib that was found to ensure it is thin.
        if dylibs:
            # Do this in a threadpool to make it run faster.
            progress_bar = self.console.progress_bar()
            self.console.info(f"Thinning libraries in {app_packages.name}...")
            task_id = progress_bar.add_task("Create fat libraries", total=len(dylibs))
            with (
                progress_bar,
                concurrent.futures.ThreadPoolExecutor(
                    max_workers=1 if self.console.is_deep_debug else None
                ) as executor,
            ):
                futures = []
                for path in dylibs:
                    future = executor.submit(
                        self.ensure_thin_binary,
                        path=path,
                        arch=arch,
                    )
                    futures.append(future)
                for future in concurrent.futures.as_completed(futures):
                    progress_bar.update(task_id, advance=1)
                    if future.exception():
                        raise future.exception()
        else:
            self.console.info("No libraries require thinning.")

    def merge_app_packages(
        self,
        target_app_packages: Path,
        sources: list[Path],
    ):
        """Merge the contents of multiple source app packages into a final location.

        :param target_app_packages: The location of the final merged app packages
        :param sources: A list of app packages folders to be merged.
        """
        # Clear any existing app packages folder.
        if target_app_packages.is_dir():
            self.tools.shutil.rmtree(target_app_packages)
        self.tools.os.mkdir(target_app_packages)

        # Copy all the non-library files from the source to the target. If a file exists
        # in multiple sources, do a checksum, and warn if there is a discrepancy. If
        # there is a discrepancy, the version from the first source is what will be
        # used. Don't warn if the file is in a __pycache__ folder; these will always be
        # different, but they'll be purged later anyway. Don't warn for anything in a
        # .dist-info folder either; these are going to be different because of platform
        # difference, but core package metadata should be consistent.
        dylibs = set()
        digests = {}
        for source_app_packages in sources:
            with self.console.wait_bar(f"Merging {source_app_packages.name}..."):
                for source_path in source_app_packages.glob("**/*"):
                    relative_path = source_path.relative_to(source_app_packages)
                    target_path = target_app_packages / relative_path
                    if source_path.is_dir():
                        target_path.mkdir(exist_ok=True)
                    else:
                        if is_mach_o_binary(source_path):
                            # Dynamic libraries need to be merged in a second pass
                            dylibs.add(relative_path)
                        elif target_path.exists():
                            # The file already exists. Check for differences; if there
                            # are any differences outside `dist-info` or `__pycache__`
                            # folders, warn the user.
                            digest = sha256_file_digest(source_path)
                            if digests[relative_path] != digest and not (
                                relative_path.parent.name == "__pycache__"
                                or Path(relative_path.parts[0]).suffix == ".dist-info"
                            ):
                                self.console.warning(
                                    f"{relative_path} has different content between"
                                    " sources; ignoring"
                                    f" {source_app_packages.suffix[1:]} version. This"
                                    " is usually safe if the file content is not used"
                                    " at runtime. See"
                                    " https://briefcase.readthedocs.io/en/stable/reference/platforms/macOS/index.html#inconsistent-content-in-non-universal-wheels"
                                    " for more details."
                                )
                        else:
                            # The file doesn't exist yet; copy it as is (including
                            # permissions), and store the digest for later comparison
                            self.tools.shutil.copy(source_path, target_path)
                            digests[relative_path] = sha256_file_digest(source_path)

        # Call lipo on each dylib that was found to create the fat version.
        if dylibs:
            # Do this in a threadpool to make it run faster.
            progress_bar = self.console.progress_bar()
            self.console.info("Merging libraries...")
            task_id = progress_bar.add_task("Create fat libraries", total=len(dylibs))
            with (
                progress_bar,
                concurrent.futures.ThreadPoolExecutor(
                    max_workers=1 if self.console.is_deep_debug else None
                ) as executor,
            ):
                futures = []
                for relative_path in dylibs:
                    future = executor.submit(
                        self.lipo_dylib,
                        target_path=target_app_packages,
                        relative_path=relative_path,
                        sources=sources,
                    )
                    futures.append(future)
                for future in concurrent.futures.as_completed(futures):
                    progress_bar.update(task_id, advance=1)
                    if future.exception():
                        raise future.exception()
        else:
            self.console.info("No libraries require merging.")


def is_uti_core_type(uti: str) -> bool:  # pragma: no-cover-if-not-macos
    """Check if a UTI is a built-in Core Type.

    This function checks if a given UTI is a built-in Core Type by reading the
    system's CoreTypes Info.plist file. If the file is not found, it assumes
    that the system is not macOS or the file has been moved in recent macOS
    versions, and returns False.

    Args:
        uti: The UTI to check.

    Returns:
        True if the UTI is a built-in Core Type, False otherwise.
    """
    try:
        plist_data = pathlib.Path(CORETYPES_PATH).read_bytes()
    except FileNotFoundError:
        # If the file is not found, we assume that the system is not macOS
        # or the file has been moved in recent macOS versions.
        # In this case, we return False to indicate that the UTI is not built-in.
        return False
    plist = plistlib.loads(plist_data)
    return uti in {
        type_declaration["UTTypeIdentifier"]
        for type_declaration in (
            plist["UTExportedTypeDeclarations"] + plist["UTImportedTypeDeclarations"]
        )
    }


def mime_type_to_uti(mime_type: str) -> str | None:  # pragma: no-cover-if-not-macos
    """Convert a MIME type to a Uniform Type Identifier (UTI).

    This function reads the system's CoreTypes Info.plist file to determine the
    UTI for a given MIME type.

    Args:
        mime_type: The MIME type to convert.

    Returns:
        The UTI for the MIME type, or None if the UTI cannot be determined.
    """
    try:
        plist_data = pathlib.Path(CORETYPES_PATH).read_bytes()
    except FileNotFoundError:
        # If the file is not found, we assume that the system is not macOS
        # or the file has been moved in recent macOS versions.
        # In this case, we return None to indicate that the UTI cannot be determined.
        return None
    plist = plistlib.loads(plist_data)
    type_declarations = (
        plist["UTExportedTypeDeclarations"] + plist["UTImportedTypeDeclarations"]
    )
    for type_declaration in type_declarations:
        # We check both the system built-in types (exported) and the known
        # third-party types (imported) to find the UTI for the given MIME type.
        # Most type declarations will have a UTTypeTagSpecification dictionary
        # with a "public.mime-type" key. That can be either a list of MIME types
        # or a single MIME type. We check if the MIME type is in the list or
        # matches the single MIME type. If we find a match, we return the UTI
        # identifier. If we don't find a match, we return None.

        mime_types = type_declaration.get("UTTypeTagSpecification", {}).get(
            "public.mime-type", []
        )
        if isinstance(mime_types, list):
            # Most MIME types are declared as a list even if they are a
            # single type. Some types define multiple closely-related MIME
            # types.
            if mime_type in mime_types:
                return type_declaration["UTTypeIdentifier"]
        else:
            # some MIME types are declared as a single type
            if mime_types == mime_type:
                return type_declaration["UTTypeIdentifier"]

    # If no match is found in the entire list, return None
    return None



from textwrap import dedent

import pytest

from briefcase.exceptions import BriefcaseCommandError

from .....utils import create_file


def test_license_file_txt(create_command, first_app_templated, tmp_path):
    """A license can be specified as a TXT file."""
    create_file(tmp_path / "base_path/LICENSE", "This is a license file")

    create_command.install_license(first_app_templated)

    # A LICENSE.rtf file has been written, in RTF format, containing the license text.
    license = tmp_path / "base_path/build/first-app/windows/app/LICENSE.rtf"
    assert license.is_file()

    license_text = license.read_text(encoding="utf-8")
    assert license_text.startswith("{\\rtf1\\ansi")
    assert "This is a license file" in license_text
    assert license_text.endswith("\n}")


def test_license_file_rtf(create_command, first_app_templated, tmp_path):
    """A license can be specified as an RTF file."""
    first_app_templated.license["file"] = "LICENSE.rtf"
    orig_license = tmp_path / "base_path/LICENSE.rtf"
    create_file(
        orig_license,
        dedent(
            """\
            {\\rtf1\\ansi\\deff0 {\\fonttbl {\\f0 Courier;}}
            This is a document that contains bullet points:\x20
            \\par\\line
            \\bullet first, something short\x20
            \\bullet then, something longer that needs to\x20
            run onto multiple lines\x20
            \\bullet last, something short again\x20
            \\par\\line
            Then a closing paragraph.\x20
            \\par\\line
            }"""
        ),
    )
    create_command.install_license(first_app_templated)

    # The license RTF file is the same as the original
    license = tmp_path / "base_path/build/first-app/windows/app/LICENSE.rtf"
    assert license.is_file()
    assert license.read_text(encoding="utf-8") == orig_license.read_text(
        encoding="utf-8"
    )


def test_missing_license_file(create_command, first_app_templated):
    """If a named license file is missing, an error is raised."""
    with pytest.raises(
        BriefcaseCommandError,
        match=r"However, this file does not exist.",
    ):
        create_command.install_license(first_app_templated)


def test_license_text(create_command, first_app_templated, tmp_path):
    """A license provided as text is converted into a file."""
    del first_app_templated.license["file"]
    first_app_templated.license["text"] = "This is a license\nDo what you want."

    create_command.install_license(first_app_templated)

    # A license file was written, and contains the licence text
    license = tmp_path / "base_path/build/first-app/windows/app/LICENSE.rtf"
    license_text = license.read_text(encoding="utf-8")
    assert license_text.startswith("{\\rtf1\\ansi")
    assert "This is a license " in license_text
    assert license_text.endswith("\n}")


def test_license_text_suspicious(create_command, first_app_templated, tmp_path, capsys):
    """A single line license text is flagged as suspicious, but converted."""
    del first_app_templated.license["file"]
    first_app_templated.license["text"] = "BSD 3 clause"

    create_command.install_license(first_app_templated)

    license = tmp_path / "base_path/build/first-app/windows/app/LICENSE.rtf"
    assert license.is_file()

    license_text = license.read_text(encoding="utf-8")
    assert license_text.startswith("{\\rtf1\\ansi")
    assert "BSD 3 clause" in license_text
    assert license_text.endswith("\n}")

    # The user was warned that the text probably isn't a full license
    console = capsys.readouterr().out
    assert "ensure that the contents of this file is adequate." in console


def test_no_license(create_command, first_app_templated):
    """If an app doesn't provide a known license definition, an error is raised."""
    del first_app_templated.license["file"]
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Your project does not contain a `license` definition.",
    ):
        create_command.install_license(first_app_templated)



from __future__ import annotations

from importlib.metadata import entry_points

from briefcase.channels.base import (
    BasePublicationChannel,
    PublishCommandAPI,  # noqa: F401
)
from briefcase.exceptions import BriefcaseCommandError


def get_publication_channels(
    platform: str,
    output_format: str,
) -> dict[str, type[BasePublicationChannel]]:
    """Load built-in and third-party publication channels for a platform/format.

    :param platform: The target platform (e.g., "iOS")
    :param output_format: The output format (e.g., "Xcode")
    :returns: A dict mapping channel names to channel classes.
    """
    return {
        ep.name: ep.load()
        for ep in entry_points(group=f"briefcase.channels.{platform}.{output_format}")
    }


def get_publication_channel(
    name: str,
    platform: str,
    output_format: str,
) -> BasePublicationChannel:
    """Get a publication channel by name for a platform/format.

    :param name: The channel name
    :param platform: The target platform
    :param output_format: The output format
    :returns: An instantiated publication channel.
    """
    channels = get_publication_channels(platform, output_format)
    if name not in channels:
        available = ", ".join(sorted(channels)) or "none"
        raise BriefcaseCommandError(
            f"Unknown publication channel: {name}\n\n"
            f"Available channels for {platform}/{output_format}: {available}"
        )
    return channels[name]()



from briefcase.bootstraps.base import BaseGuiBootstrap


def test_base_bootstrap_fields():
    """Defined fields are implemented and implemented fields are defined."""
    for field in BaseGuiBootstrap.fields:
        assert field in BaseGuiBootstrap.__dict__

    for field in [
        attr
        for attr in BaseGuiBootstrap.__dict__
        if not attr.startswith("_")
        and attr
        not in {
            "fields",
            "display_name_annotation",
            "extra_context",
            "post_generate",
        }
    ]:
        assert field in BaseGuiBootstrap.fields



import sys
from pathlib import Path


def test_creates_marker_when_missing(noop_context):
    """A marker file is created when it doesn't exist."""
    assert not noop_context.marker_path.exists()

    noop_context.create()

    assert noop_context.created is True
    assert noop_context.marker_path.exists()
    assert noop_context.marker_path.read_text(encoding="utf-8") == sys.executable


def test_returns_false_when_marker_unchanged(noop_context):
    """Environment is not marked as created when executable hasn't changed."""
    noop_context.marker_path.parent.mkdir(parents=True, exist_ok=True)
    noop_context.marker_path.write_text(sys.executable, encoding="utf-8")

    noop_context.create()

    assert noop_context.created is False
    assert noop_context.marker_path.read_text(encoding="utf-8") == sys.executable


def test_updates_marker_when_executable_changed(noop_context):
    """Create updates marker when executable changes."""
    noop_context.marker_path.parent.mkdir(parents=True, exist_ok=True)
    old_executable = "/old/python/path"
    noop_context.marker_path.write_text(old_executable, encoding="utf-8")

    noop_context.create()

    assert noop_context.created is True
    assert noop_context.marker_path.read_text(encoding="utf-8") == sys.executable
    assert noop_context.marker_path.read_text(encoding="utf-8") != old_executable


def test_handles_whitespace_in_marker(noop_context):
    """Create strips whitespace when reading marker."""
    noop_context.marker_path.parent.mkdir(parents=True, exist_ok=True)
    noop_context.marker_path.write_text(f"  {sys.executable}  \n", encoding="utf-8")

    noop_context.create()

    assert noop_context.created is False


def test_handles_oserror_during_read(noop_context, monkeypatch):
    """Create handles OSError by recreating marker."""
    noop_context.marker_path.parent.mkdir(parents=True, exist_ok=True)
    noop_context.marker_path.write_text("something", encoding="utf-8")

    original_read = Path.read_text

    def mock_read_text(*args, **kwargs):
        raise OSError("Permission denied")

    monkeypatch.setattr(Path, "read_text", mock_read_text)

    noop_context.create()

    assert noop_context.created is True

    monkeypatch.setattr(Path, "read_text", original_read)
    assert noop_context.marker_path.read_text(encoding="utf-8") == sys.executable


def test_handles_unicode_decode_error(noop_context):
    """Create handles UnicodeDecodeError by recreating marker."""
    noop_context.marker_path.parent.mkdir(parents=True, exist_ok=True)
    noop_context.marker_path.write_bytes(b"\xff\xfe\xfd")

    noop_context.create()

    assert noop_context.created is True
    assert noop_context.marker_path.read_text(encoding="utf-8") == sys.executable


def test_marker_persists_across_instances(mock_tools, tmp_path):
    """Marker file persists and is readable by new instances."""
    from briefcase.integrations.virtual_environment import NoOpVenvContext

    venv_path = tmp_path / "venv"

    context1 = NoOpVenvContext(tools=mock_tools, venv_path=venv_path)
    context1.create()
    assert context1.created is True

    context2 = NoOpVenvContext(tools=mock_tools, venv_path=venv_path)
    context2.create()
    assert context2.created is False



import json
import os
import sys
from unittest.mock import MagicMock

import briefcase_debugger
import debugpy
import pytest
from briefcase_debugger.config import AppPathMappings


def test_no_env_vars(monkeypatch, capsys):
    """Nothing happens, when no env vars are set."""
    os_environ = {}
    monkeypatch.setattr(os, "environ", os_environ)

    # start test function
    briefcase_debugger.start_remote_debugger()

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == ""


def test_no_debugger_verbose(monkeypatch, capsys):
    """Nothing happens except a short message, when only verbose is requested."""
    os_environ = {}
    os_environ["BRIEFCASE_DEBUG"] = "1"
    monkeypatch.setattr(os, "environ", os_environ)

    # start test function
    briefcase_debugger.start_remote_debugger()

    captured = capsys.readouterr()
    assert (
        captured.out
        == "No 'BRIEFCASE_DEBUGGER' environment variable found. Debugger not starting.\n"
    )
    assert captured.err == ""


@pytest.mark.parametrize(
    ("os_name", "app_path_mappings", "sys_path", "expected_path_mappings"),
    [
        (
            "nt",
            AppPathMappings(
                device_sys_path_regex="app$",
                device_subfolders=["helloworld"],
                host_folders=["C:\\PROJECT_ROOT\\src\\helloworld"],
            ),
            ["C:\\PROJECT_ROOT\\build\\helloworld\\windows\\app\\src\\app"],
            [
                (
                    "C:\\PROJECT_ROOT\\src\\helloworld",
                    "C:\\PROJECT_ROOT\\build\\helloworld\\windows\\app\\src\\app\\helloworld",
                )
            ],
        ),
        (
            "posix",
            AppPathMappings(
                device_sys_path_regex="app$",
                device_subfolders=["helloworld"],
                host_folders=["/PROJECT_ROOT/src/helloworld"],
            ),
            [
                "/PROJECT_ROOT/build/helloworld/macos/app/Hello World.app/Contents/Resources/app"
            ],
            [
                (
                    "/PROJECT_ROOT/src/helloworld",
                    "/PROJECT_ROOT/build/helloworld/macos/app/Hello World.app/Contents/Resources/app/helloworld",
                )
            ],
        ),
    ],
)
@pytest.mark.parametrize(
    ("verbose", "some_verbose_output", "pydevd_trace_level"),
    [
        (True, "Extracted path mappings:\n[0] host =   ", 3),
        (False, "", 0),
    ],
)
def test_with_debugger(
    os_name: str,
    app_path_mappings: AppPathMappings,
    sys_path: list[str],
    expected_path_mappings: list[tuple[str, str]],
    verbose: bool,
    some_verbose_output: str,
    pydevd_trace_level: int,
    monkeypatch,
    capsys,
):
    """Normal debug session."""
    if os.name != os_name:
        pytest.skip(f"Test only runs on {os_name} systems")

    os_environ = {}
    os_environ["BRIEFCASE_DEBUG"] = "1" if verbose else "0"
    os_environ["BRIEFCASE_DEBUGGER"] = json.dumps(
        {
            "debugger": "debugpy",
            "host": "somehost",
            "port": 9999,
            "host_os": "SomeOS",
            "app_path_mappings": app_path_mappings,
            "app_packages_path_mappings": None,
        }
    )
    monkeypatch.setattr(os, "environ", os_environ)

    monkeypatch.setattr(sys, "path", sys_path)

    fake_debugpy_listen = MagicMock()
    monkeypatch.setattr(debugpy, "listen", fake_debugpy_listen)

    fake_debugpy_wait_for_client = MagicMock()
    monkeypatch.setattr(debugpy, "wait_for_client", fake_debugpy_wait_for_client)

    # pydevd is dynamically loaded and only available when a real debugger is attached. So
    # we fake the whole module, as otherwise the import in start_remote_debugger would fail
    fake_pydevd = MagicMock()
    monkeypatch.setitem(sys.modules, "pydevd", fake_pydevd)
    fake_pydevd.DebugInfoHolder.DEBUG_TRACE_LEVEL = 0
    fake_pydevd_file_utils = MagicMock()
    fake_pydevd_file_utils.setup_client_server_paths.return_value = None
    monkeypatch.setitem(sys.modules, "pydevd_file_utils", fake_pydevd_file_utils)

    # start test function
    briefcase_debugger.start_remote_debugger()

    fake_debugpy_listen.assert_called_once_with(
        ("somehost", 9999),
        in_process_debug_adapter=True,
    )

    fake_debugpy_wait_for_client.assert_called_once()
    fake_pydevd_file_utils.setup_client_server_paths.assert_called_once_with(
        expected_path_mappings
    )

    captured = capsys.readouterr()
    assert "Waiting for debugger to attach..." in captured.out
    assert captured.err == ""

    assert some_verbose_output in captured.out
    assert pydevd_trace_level == fake_pydevd.DebugInfoHolder.DEBUG_TRACE_LEVEL


def test_with_debugger_without_path_mappings(monkeypatch, capsys):
    """Debug session without path mappings."""
    os_environ = {}
    os_environ["BRIEFCASE_DEBUG"] = "0"
    os_environ["BRIEFCASE_DEBUGGER"] = json.dumps(
        {
            "debugger": "debugpy",
            "host": "somehost",
            "port": 9999,
            "host_os": "SomeOS",
            "app_path_mappings": None,
            "app_packages_path_mappings": None,
        }
    )
    monkeypatch.setattr(os, "environ", os_environ)

    fake_debugpy_listen = MagicMock()
    monkeypatch.setattr(debugpy, "listen", fake_debugpy_listen)

    fake_debugpy_wait_for_client = MagicMock()
    monkeypatch.setattr(debugpy, "wait_for_client", fake_debugpy_wait_for_client)

    # pydevd is dynamically loaded and only available when a real debugger is attached. So
    # we fake the whole module, as otherwise the import in start_remote_debugger would fail
    fake_pydevd = MagicMock()
    monkeypatch.setitem(sys.modules, "pydevd", fake_pydevd)
    fake_pydevd.DebugInfoHolder.DEBUG_TRACE_LEVEL = 0
    fake_pydevd_file_utils = MagicMock()
    fake_pydevd_file_utils.setup_client_server_paths.return_value = None
    monkeypatch.setitem(sys.modules, "pydevd_file_utils", fake_pydevd_file_utils)

    # start test function
    briefcase_debugger.start_remote_debugger()

    fake_debugpy_listen.assert_called_once_with(
        ("somehost", 9999),
        in_process_debug_adapter=True,
    )

    fake_debugpy_wait_for_client.assert_called_once()
    fake_pydevd_file_utils.setup_client_server_paths.assert_not_called()

    captured = capsys.readouterr()
    assert "Waiting for debugger to attach..." in captured.out
    assert captured.err == ""
